import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from .. import types


class BankJSON(typing.TypedDict):
    group: str
    name: list[int]
    mint: str
    vault: str
    oracle: str
    oracle_config: types.oracle_config.OracleConfigJSON
    stable_price_model: types.stable_price_model.StablePriceModelJSON
    deposit_index: types.i80f48.I80F48JSON
    borrow_index: types.i80f48.I80F48JSON
    indexed_deposits: types.i80f48.I80F48JSON
    indexed_borrows: types.i80f48.I80F48JSON
    index_last_updated: int
    bank_rate_last_updated: int
    avg_utilization: types.i80f48.I80F48JSON
    adjustment_factor: types.i80f48.I80F48JSON
    util0: types.i80f48.I80F48JSON
    rate0: types.i80f48.I80F48JSON
    util1: types.i80f48.I80F48JSON
    rate1: types.i80f48.I80F48JSON
    max_rate: types.i80f48.I80F48JSON
    collected_fees_native: types.i80f48.I80F48JSON
    loan_origination_fee_rate: types.i80f48.I80F48JSON
    loan_fee_rate: types.i80f48.I80F48JSON
    maint_asset_weight: types.i80f48.I80F48JSON
    init_asset_weight: types.i80f48.I80F48JSON
    maint_liab_weight: types.i80f48.I80F48JSON
    init_liab_weight: types.i80f48.I80F48JSON
    liquidation_fee: types.i80f48.I80F48JSON
    dust: types.i80f48.I80F48JSON
    flash_loan_token_account_initial: int
    flash_loan_approved_amount: int
    token_index: int
    bump: int
    mint_decimals: int
    bank_num: int
    min_vault_to_deposits_ratio: float
    net_borrow_limit_window_size_ts: int
    last_net_borrows_window_start_ts: int
    net_borrow_limit_per_window_quote: int
    net_borrows_in_window: int
    borrow_weight_scale_start_quote: float
    deposit_weight_scale_start_quote: float
    reduce_only: int
    reserved: list[int]


@dataclass
class Bank:
    discriminator: typing.ClassVar = b"\x8e1\xa6\xf22Ba\xbc"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "name" / borsh.U8[16],
        "mint" / BorshPubkey,
        "vault" / BorshPubkey,
        "oracle" / BorshPubkey,
        "oracle_config" / types.oracle_config.OracleConfig.layout,
        "stable_price_model" / types.stable_price_model.StablePriceModel.layout,
        "deposit_index" / types.i80f48.I80F48.layout,
        "borrow_index" / types.i80f48.I80F48.layout,
        "indexed_deposits" / types.i80f48.I80F48.layout,
        "indexed_borrows" / types.i80f48.I80F48.layout,
        "index_last_updated" / borsh.U64,
        "bank_rate_last_updated" / borsh.U64,
        "avg_utilization" / types.i80f48.I80F48.layout,
        "adjustment_factor" / types.i80f48.I80F48.layout,
        "util0" / types.i80f48.I80F48.layout,
        "rate0" / types.i80f48.I80F48.layout,
        "util1" / types.i80f48.I80F48.layout,
        "rate1" / types.i80f48.I80F48.layout,
        "max_rate" / types.i80f48.I80F48.layout,
        "collected_fees_native" / types.i80f48.I80F48.layout,
        "loan_origination_fee_rate" / types.i80f48.I80F48.layout,
        "loan_fee_rate" / types.i80f48.I80F48.layout,
        "maint_asset_weight" / types.i80f48.I80F48.layout,
        "init_asset_weight" / types.i80f48.I80F48.layout,
        "maint_liab_weight" / types.i80f48.I80F48.layout,
        "init_liab_weight" / types.i80f48.I80F48.layout,
        "liquidation_fee" / types.i80f48.I80F48.layout,
        "dust" / types.i80f48.I80F48.layout,
        "flash_loan_token_account_initial" / borsh.U64,
        "flash_loan_approved_amount" / borsh.U64,
        "token_index" / borsh.U16,
        "bump" / borsh.U8,
        "mint_decimals" / borsh.U8,
        "bank_num" / borsh.U32,
        "min_vault_to_deposits_ratio" / borsh.F64,
        "net_borrow_limit_window_size_ts" / borsh.U64,
        "last_net_borrows_window_start_ts" / borsh.U64,
        "net_borrow_limit_per_window_quote" / borsh.I64,
        "net_borrows_in_window" / borsh.I64,
        "borrow_weight_scale_start_quote" / borsh.F64,
        "deposit_weight_scale_start_quote" / borsh.F64,
        "reduce_only" / borsh.U8,
        "reserved" / borsh.U8[2119],
    )
    group: PublicKey
    name: list[int]
    mint: PublicKey
    vault: PublicKey
    oracle: PublicKey
    oracle_config: types.oracle_config.OracleConfig
    stable_price_model: types.stable_price_model.StablePriceModel
    deposit_index: types.i80f48.I80F48
    borrow_index: types.i80f48.I80F48
    indexed_deposits: types.i80f48.I80F48
    indexed_borrows: types.i80f48.I80F48
    index_last_updated: int
    bank_rate_last_updated: int
    avg_utilization: types.i80f48.I80F48
    adjustment_factor: types.i80f48.I80F48
    util0: types.i80f48.I80F48
    rate0: types.i80f48.I80F48
    util1: types.i80f48.I80F48
    rate1: types.i80f48.I80F48
    max_rate: types.i80f48.I80F48
    collected_fees_native: types.i80f48.I80F48
    loan_origination_fee_rate: types.i80f48.I80F48
    loan_fee_rate: types.i80f48.I80F48
    maint_asset_weight: types.i80f48.I80F48
    init_asset_weight: types.i80f48.I80F48
    maint_liab_weight: types.i80f48.I80F48
    init_liab_weight: types.i80f48.I80F48
    liquidation_fee: types.i80f48.I80F48
    dust: types.i80f48.I80F48
    flash_loan_token_account_initial: int
    flash_loan_approved_amount: int
    token_index: int
    bump: int
    mint_decimals: int
    bank_num: int
    min_vault_to_deposits_ratio: float
    net_borrow_limit_window_size_ts: int
    last_net_borrows_window_start_ts: int
    net_borrow_limit_per_window_quote: int
    net_borrows_in_window: int
    borrow_weight_scale_start_quote: float
    deposit_weight_scale_start_quote: float
    reduce_only: int
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["Bank"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Bank"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Bank"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Bank":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Bank.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            group=dec.group,
            name=dec.name,
            mint=dec.mint,
            vault=dec.vault,
            oracle=dec.oracle,
            oracle_config=types.oracle_config.OracleConfig.from_decoded(
                dec.oracle_config
            ),
            stable_price_model=types.stable_price_model.StablePriceModel.from_decoded(
                dec.stable_price_model
            ),
            deposit_index=types.i80f48.I80F48.from_decoded(dec.deposit_index),
            borrow_index=types.i80f48.I80F48.from_decoded(dec.borrow_index),
            indexed_deposits=types.i80f48.I80F48.from_decoded(dec.indexed_deposits),
            indexed_borrows=types.i80f48.I80F48.from_decoded(dec.indexed_borrows),
            index_last_updated=dec.index_last_updated,
            bank_rate_last_updated=dec.bank_rate_last_updated,
            avg_utilization=types.i80f48.I80F48.from_decoded(dec.avg_utilization),
            adjustment_factor=types.i80f48.I80F48.from_decoded(dec.adjustment_factor),
            util0=types.i80f48.I80F48.from_decoded(dec.util0),
            rate0=types.i80f48.I80F48.from_decoded(dec.rate0),
            util1=types.i80f48.I80F48.from_decoded(dec.util1),
            rate1=types.i80f48.I80F48.from_decoded(dec.rate1),
            max_rate=types.i80f48.I80F48.from_decoded(dec.max_rate),
            collected_fees_native=types.i80f48.I80F48.from_decoded(
                dec.collected_fees_native
            ),
            loan_origination_fee_rate=types.i80f48.I80F48.from_decoded(
                dec.loan_origination_fee_rate
            ),
            loan_fee_rate=types.i80f48.I80F48.from_decoded(dec.loan_fee_rate),
            maint_asset_weight=types.i80f48.I80F48.from_decoded(dec.maint_asset_weight),
            init_asset_weight=types.i80f48.I80F48.from_decoded(dec.init_asset_weight),
            maint_liab_weight=types.i80f48.I80F48.from_decoded(dec.maint_liab_weight),
            init_liab_weight=types.i80f48.I80F48.from_decoded(dec.init_liab_weight),
            liquidation_fee=types.i80f48.I80F48.from_decoded(dec.liquidation_fee),
            dust=types.i80f48.I80F48.from_decoded(dec.dust),
            flash_loan_token_account_initial=dec.flash_loan_token_account_initial,
            flash_loan_approved_amount=dec.flash_loan_approved_amount,
            token_index=dec.token_index,
            bump=dec.bump,
            mint_decimals=dec.mint_decimals,
            bank_num=dec.bank_num,
            min_vault_to_deposits_ratio=dec.min_vault_to_deposits_ratio,
            net_borrow_limit_window_size_ts=dec.net_borrow_limit_window_size_ts,
            last_net_borrows_window_start_ts=dec.last_net_borrows_window_start_ts,
            net_borrow_limit_per_window_quote=dec.net_borrow_limit_per_window_quote,
            net_borrows_in_window=dec.net_borrows_in_window,
            borrow_weight_scale_start_quote=dec.borrow_weight_scale_start_quote,
            deposit_weight_scale_start_quote=dec.deposit_weight_scale_start_quote,
            reduce_only=dec.reduce_only,
            reserved=dec.reserved,
        )

    def to_json(self) -> BankJSON:
        return {
            "group": str(self.group),
            "name": self.name,
            "mint": str(self.mint),
            "vault": str(self.vault),
            "oracle": str(self.oracle),
            "oracle_config": self.oracle_config.to_json(),
            "stable_price_model": self.stable_price_model.to_json(),
            "deposit_index": self.deposit_index.to_json(),
            "borrow_index": self.borrow_index.to_json(),
            "indexed_deposits": self.indexed_deposits.to_json(),
            "indexed_borrows": self.indexed_borrows.to_json(),
            "index_last_updated": self.index_last_updated,
            "bank_rate_last_updated": self.bank_rate_last_updated,
            "avg_utilization": self.avg_utilization.to_json(),
            "adjustment_factor": self.adjustment_factor.to_json(),
            "util0": self.util0.to_json(),
            "rate0": self.rate0.to_json(),
            "util1": self.util1.to_json(),
            "rate1": self.rate1.to_json(),
            "max_rate": self.max_rate.to_json(),
            "collected_fees_native": self.collected_fees_native.to_json(),
            "loan_origination_fee_rate": self.loan_origination_fee_rate.to_json(),
            "loan_fee_rate": self.loan_fee_rate.to_json(),
            "maint_asset_weight": self.maint_asset_weight.to_json(),
            "init_asset_weight": self.init_asset_weight.to_json(),
            "maint_liab_weight": self.maint_liab_weight.to_json(),
            "init_liab_weight": self.init_liab_weight.to_json(),
            "liquidation_fee": self.liquidation_fee.to_json(),
            "dust": self.dust.to_json(),
            "flash_loan_token_account_initial": self.flash_loan_token_account_initial,
            "flash_loan_approved_amount": self.flash_loan_approved_amount,
            "token_index": self.token_index,
            "bump": self.bump,
            "mint_decimals": self.mint_decimals,
            "bank_num": self.bank_num,
            "min_vault_to_deposits_ratio": self.min_vault_to_deposits_ratio,
            "net_borrow_limit_window_size_ts": self.net_borrow_limit_window_size_ts,
            "last_net_borrows_window_start_ts": self.last_net_borrows_window_start_ts,
            "net_borrow_limit_per_window_quote": self.net_borrow_limit_per_window_quote,
            "net_borrows_in_window": self.net_borrows_in_window,
            "borrow_weight_scale_start_quote": self.borrow_weight_scale_start_quote,
            "deposit_weight_scale_start_quote": self.deposit_weight_scale_start_quote,
            "reduce_only": self.reduce_only,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: BankJSON) -> "Bank":
        return cls(
            group=PublicKey(obj["group"]),
            name=obj["name"],
            mint=PublicKey(obj["mint"]),
            vault=PublicKey(obj["vault"]),
            oracle=PublicKey(obj["oracle"]),
            oracle_config=types.oracle_config.OracleConfig.from_json(
                obj["oracle_config"]
            ),
            stable_price_model=types.stable_price_model.StablePriceModel.from_json(
                obj["stable_price_model"]
            ),
            deposit_index=types.i80f48.I80F48.from_json(obj["deposit_index"]),
            borrow_index=types.i80f48.I80F48.from_json(obj["borrow_index"]),
            indexed_deposits=types.i80f48.I80F48.from_json(obj["indexed_deposits"]),
            indexed_borrows=types.i80f48.I80F48.from_json(obj["indexed_borrows"]),
            index_last_updated=obj["index_last_updated"],
            bank_rate_last_updated=obj["bank_rate_last_updated"],
            avg_utilization=types.i80f48.I80F48.from_json(obj["avg_utilization"]),
            adjustment_factor=types.i80f48.I80F48.from_json(obj["adjustment_factor"]),
            util0=types.i80f48.I80F48.from_json(obj["util0"]),
            rate0=types.i80f48.I80F48.from_json(obj["rate0"]),
            util1=types.i80f48.I80F48.from_json(obj["util1"]),
            rate1=types.i80f48.I80F48.from_json(obj["rate1"]),
            max_rate=types.i80f48.I80F48.from_json(obj["max_rate"]),
            collected_fees_native=types.i80f48.I80F48.from_json(
                obj["collected_fees_native"]
            ),
            loan_origination_fee_rate=types.i80f48.I80F48.from_json(
                obj["loan_origination_fee_rate"]
            ),
            loan_fee_rate=types.i80f48.I80F48.from_json(obj["loan_fee_rate"]),
            maint_asset_weight=types.i80f48.I80F48.from_json(obj["maint_asset_weight"]),
            init_asset_weight=types.i80f48.I80F48.from_json(obj["init_asset_weight"]),
            maint_liab_weight=types.i80f48.I80F48.from_json(obj["maint_liab_weight"]),
            init_liab_weight=types.i80f48.I80F48.from_json(obj["init_liab_weight"]),
            liquidation_fee=types.i80f48.I80F48.from_json(obj["liquidation_fee"]),
            dust=types.i80f48.I80F48.from_json(obj["dust"]),
            flash_loan_token_account_initial=obj["flash_loan_token_account_initial"],
            flash_loan_approved_amount=obj["flash_loan_approved_amount"],
            token_index=obj["token_index"],
            bump=obj["bump"],
            mint_decimals=obj["mint_decimals"],
            bank_num=obj["bank_num"],
            min_vault_to_deposits_ratio=obj["min_vault_to_deposits_ratio"],
            net_borrow_limit_window_size_ts=obj["net_borrow_limit_window_size_ts"],
            last_net_borrows_window_start_ts=obj["last_net_borrows_window_start_ts"],
            net_borrow_limit_per_window_quote=obj["net_borrow_limit_per_window_quote"],
            net_borrows_in_window=obj["net_borrows_in_window"],
            borrow_weight_scale_start_quote=obj["borrow_weight_scale_start_quote"],
            deposit_weight_scale_start_quote=obj["deposit_weight_scale_start_quote"],
            reduce_only=obj["reduce_only"],
            reserved=obj["reserved"],
        )



from mango_explorer_v4.accounts.mango_account import MangoAccount
from mango_explorer_v4.types.token_position import TokenPosition
from mango_explorer_v4.types.serum3_orders import Serum3Orders
from mango_explorer_v4.types.perp_position import PerpPosition
from mango_explorer_v4.helpers.perp_position import PerpPositionHelper
from mango_explorer_v4.helpers.serum3_orders import Serum3OrdersHelper
from mango_explorer_v4.helpers.token_position import TokenPositionHelper


class MangoAccountHelper:
    @staticmethod
    def active_token_positions(mango_account: MangoAccount) -> [TokenPosition]:
        return [
            token_position for token_position in mango_account.tokens
            if TokenPositionHelper.is_active(token_position)
        ]

    @staticmethod
    def active_serum3_orders(mango_account: MangoAccount) -> [Serum3Orders]:
        return [
            serum3_orders for serum3_orders in mango_account.serum3
            if Serum3OrdersHelper.is_active(serum3_orders)
        ]

    @staticmethod
    def active_perp_positions(mango_account: MangoAccount) -> [PerpPosition]:
        return [
            perp_position for perp_position in mango_account.perps
            if PerpPositionHelper.is_active(perp_position)
        ]



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from ..program_id import PROGRAM_ID
from .. import types


class BookSideJSON(typing.TypedDict):
    roots: list[types.order_tree_root.OrderTreeRootJSON]
    reserved_roots: list[types.order_tree_root.OrderTreeRootJSON]
    reserved: list[int]
    nodes: types.order_tree_nodes.OrderTreeNodesJSON


@dataclass
class BookSide:
    discriminator: typing.ClassVar = b"H,\xe1\x8d\xb2\x82a9"
    layout: typing.ClassVar = borsh.CStruct(
        "roots" / types.order_tree_root.OrderTreeRoot.layout[2],
        "reserved_roots" / types.order_tree_root.OrderTreeRoot.layout[4],
        "reserved" / borsh.U8[256],
        "nodes" / types.order_tree_nodes.OrderTreeNodes.layout,
    )
    roots: list[types.order_tree_root.OrderTreeRoot]
    reserved_roots: list[types.order_tree_root.OrderTreeRoot]
    reserved: list[int]
    nodes: types.order_tree_nodes.OrderTreeNodes

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["BookSide"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["BookSide"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["BookSide"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "BookSide":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = BookSide.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            roots=list(
                map(
                    lambda item: types.order_tree_root.OrderTreeRoot.from_decoded(item),
                    dec.roots,
                )
            ),
            reserved_roots=list(
                map(
                    lambda item: types.order_tree_root.OrderTreeRoot.from_decoded(item),
                    dec.reserved_roots,
                )
            ),
            reserved=dec.reserved,
            nodes=types.order_tree_nodes.OrderTreeNodes.from_decoded(dec.nodes),
        )

    def to_json(self) -> BookSideJSON:
        return {
            "roots": list(map(lambda item: item.to_json(), self.roots)),
            "reserved_roots": list(
                map(lambda item: item.to_json(), self.reserved_roots)
            ),
            "reserved": self.reserved,
            "nodes": self.nodes.to_json(),
        }

    @classmethod
    def from_json(cls, obj: BookSideJSON) -> "BookSide":
        return cls(
            roots=list(
                map(
                    lambda item: types.order_tree_root.OrderTreeRoot.from_json(item),
                    obj["roots"],
                )
            ),
            reserved_roots=list(
                map(
                    lambda item: types.order_tree_root.OrderTreeRoot.from_json(item),
                    obj["reserved_roots"],
                )
            ),
            reserved=obj["reserved"],
            nodes=types.order_tree_nodes.OrderTreeNodes.from_json(obj["nodes"]),
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh
import decimal
import math


class I80F48JSON(typing.TypedDict):
    val: int


@dataclass
class I80F48:
    layout: typing.ClassVar = borsh.CStruct("val" / borsh.I128)
    val: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "I80F48":
        return cls(val=obj.val)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"val": self.val}

    def to_json(self) -> I80F48JSON:
        return {"val": self.val}

    @classmethod
    def from_json(cls, obj: I80F48JSON) -> "I80F48":
        return cls(val=obj["val"])

    def to_decimal(self) -> decimal.Decimal:
        return decimal.Decimal(self.val) / decimal.Decimal(2 ** 48)

    @classmethod
    def from_decimal(cls, val: decimal.Decimal) -> "I80F48":
        return cls(val=int(decimal.Decimal(val) * decimal.Decimal(2 ** 48)))



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class IxGateSetArgs(typing.TypedDict):
    ix_gate: int


layout = borsh.CStruct("ix_gate" / borsh.U128)


class IxGateSetAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey


def ix_gate_set(
    args: IxGateSetArgs,
    accounts: IxGateSetAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xc9\xb1Ub/\xdd\x8a\xf8"
    encoded_args = layout.build(
        {
            "ix_gate": args["ix_gate"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    token_info,
    serum3_info,
    perp_info,
)
import typing
from dataclasses import dataclass
from construct import Container, Construct
import borsh_construct as borsh


class HealthCacheJSON(typing.TypedDict):
    token_infos: list[token_info.TokenInfoJSON]
    serum3_infos: list[serum3_info.Serum3InfoJSON]
    perp_infos: list[perp_info.PerpInfoJSON]
    being_liquidated: bool


@dataclass
class HealthCache:
    layout: typing.ClassVar = borsh.CStruct(
        "token_infos" / borsh.Vec(typing.cast(Construct, token_info.TokenInfo.layout)),
        "serum3_infos"
        / borsh.Vec(typing.cast(Construct, serum3_info.Serum3Info.layout)),
        "perp_infos" / borsh.Vec(typing.cast(Construct, perp_info.PerpInfo.layout)),
        "being_liquidated" / borsh.Bool,
    )
    token_infos: list[token_info.TokenInfo]
    serum3_infos: list[serum3_info.Serum3Info]
    perp_infos: list[perp_info.PerpInfo]
    being_liquidated: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "HealthCache":
        return cls(
            token_infos=list(
                map(
                    lambda item: token_info.TokenInfo.from_decoded(item),
                    obj.token_infos,
                )
            ),
            serum3_infos=list(
                map(
                    lambda item: serum3_info.Serum3Info.from_decoded(item),
                    obj.serum3_infos,
                )
            ),
            perp_infos=list(
                map(lambda item: perp_info.PerpInfo.from_decoded(item), obj.perp_infos)
            ),
            being_liquidated=obj.being_liquidated,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "token_infos": list(
                map(lambda item: item.to_encodable(), self.token_infos)
            ),
            "serum3_infos": list(
                map(lambda item: item.to_encodable(), self.serum3_infos)
            ),
            "perp_infos": list(map(lambda item: item.to_encodable(), self.perp_infos)),
            "being_liquidated": self.being_liquidated,
        }

    def to_json(self) -> HealthCacheJSON:
        return {
            "token_infos": list(map(lambda item: item.to_json(), self.token_infos)),
            "serum3_infos": list(map(lambda item: item.to_json(), self.serum3_infos)),
            "perp_infos": list(map(lambda item: item.to_json(), self.perp_infos)),
            "being_liquidated": self.being_liquidated,
        }

    @classmethod
    def from_json(cls, obj: HealthCacheJSON) -> "HealthCache":
        return cls(
            token_infos=list(
                map(
                    lambda item: token_info.TokenInfo.from_json(item),
                    obj["token_infos"],
                )
            ),
            serum3_infos=list(
                map(
                    lambda item: serum3_info.Serum3Info.from_json(item),
                    obj["serum3_infos"],
                )
            ),
            perp_infos=list(
                map(lambda item: perp_info.PerpInfo.from_json(item), obj["perp_infos"])
            ),
            being_liquidated=obj["being_liquidated"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import MANGO_PROGRAM_ID


class PerpLiqBankruptcyArgs(typing.TypedDict):
    max_liab_transfer: int


layout = borsh.CStruct("max_liab_transfer" / borsh.U64)


class PerpLiqBankruptcyAccounts(typing.TypedDict):
    group: PublicKey
    perp_market: PublicKey
    liqor: PublicKey
    liqor_owner: PublicKey
    liqee: PublicKey
    settle_bank: PublicKey
    settle_vault: PublicKey
    settle_oracle: PublicKey
    insurance_vault: PublicKey


def perp_liq_bankruptcy(
    args: PerpLiqBankruptcyArgs,
    accounts: PerpLiqBankruptcyAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["liqor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["liqor_owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["liqee"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["settle_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["settle_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["settle_oracle"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x8b\xcc\xdd\x02\xe9\x05\xb5|"
    encoded_args = layout.build(
        {
            "max_liab_transfer": args["max_liab_transfer"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class OracleConfigJSON(typing.TypedDict):
    conf_filter: i80f48.I80F48JSON
    max_staleness_slots: int
    reserved: list[int]


@dataclass
class OracleConfig:
    layout: typing.ClassVar = borsh.CStruct(
        "conf_filter" / i80f48.I80F48.layout,
        "max_staleness_slots" / borsh.I64,
        "reserved" / borsh.U8[72],
    )
    conf_filter: i80f48.I80F48
    max_staleness_slots: int
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OracleConfig":
        return cls(
            conf_filter=i80f48.I80F48.from_decoded(obj.conf_filter),
            max_staleness_slots=obj.max_staleness_slots,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "conf_filter": self.conf_filter.to_encodable(),
            "max_staleness_slots": self.max_staleness_slots,
            "reserved": self.reserved,
        }

    def to_json(self) -> OracleConfigJSON:
        return {
            "conf_filter": self.conf_filter.to_json(),
            "max_staleness_slots": self.max_staleness_slots,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: OracleConfigJSON) -> "OracleConfig":
        return cls(
            conf_filter=i80f48.I80F48.from_json(obj["conf_filter"]),
            max_staleness_slots=obj["max_staleness_slots"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class HealthRegionEndAccounts(typing.TypedDict):
    account: PublicKey


def health_region_end(
    accounts: HealthRegionEndAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True)
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x1a\xb4\xeap:A\x08\xf6"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class OracleConfigParamsJSON(typing.TypedDict):
    conf_filter: float
    max_staleness_slots: typing.Optional[int]


@dataclass
class OracleConfigParams:
    layout: typing.ClassVar = borsh.CStruct(
        "conf_filter" / borsh.F32, "max_staleness_slots" / borsh.Option(borsh.U32)
    )
    conf_filter: float
    max_staleness_slots: typing.Optional[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OracleConfigParams":
        return cls(
            conf_filter=obj.conf_filter, max_staleness_slots=obj.max_staleness_slots
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "conf_filter": self.conf_filter,
            "max_staleness_slots": self.max_staleness_slots,
        }

    def to_json(self) -> OracleConfigParamsJSON:
        return {
            "conf_filter": self.conf_filter,
            "max_staleness_slots": self.max_staleness_slots,
        }

    @classmethod
    def from_json(cls, obj: OracleConfigParamsJSON) -> "OracleConfigParams":
        return cls(
            conf_filter=obj["conf_filter"],
            max_staleness_slots=obj["max_staleness_slots"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class InitJSON(typing.TypedDict):
    kind: typing.Literal["Init"]


class MaintJSON(typing.TypedDict):
    kind: typing.Literal["Maint"]


class LiquidationEndJSON(typing.TypedDict):
    kind: typing.Literal["LiquidationEnd"]


@dataclass
class Init:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Init"

    @classmethod
    def to_json(cls) -> InitJSON:
        return InitJSON(
            kind="Init",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Init": {},
        }


@dataclass
class Maint:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Maint"

    @classmethod
    def to_json(cls) -> MaintJSON:
        return MaintJSON(
            kind="Maint",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Maint": {},
        }


@dataclass
class LiquidationEnd:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "LiquidationEnd"

    @classmethod
    def to_json(cls) -> LiquidationEndJSON:
        return LiquidationEndJSON(
            kind="LiquidationEnd",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LiquidationEnd": {},
        }


HealthTypeKind = typing.Union[Init, Maint, LiquidationEnd]
HealthTypeJSON = typing.Union[InitJSON, MaintJSON, LiquidationEndJSON]


def from_decoded(obj: dict) -> HealthTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Init" in obj:
        return Init()
    if "Maint" in obj:
        return Maint()
    if "LiquidationEnd" in obj:
        return LiquidationEnd()
    raise ValueError("Invalid enum object")


def from_json(obj: HealthTypeJSON) -> HealthTypeKind:
    if obj["kind"] == "Init":
        return Init()
    if obj["kind"] == "Maint":
        return Maint()
    if obj["kind"] == "LiquidationEnd":
        return LiquidationEnd()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Init" / borsh.CStruct(),
    "Maint" / borsh.CStruct(),
    "LiquidationEnd" / borsh.CStruct(),
)



import typing
from dataclasses import dataclass
from construct import Construct
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from .. import types


class MangoAccountJSON(typing.TypedDict):
    group: str
    owner: str
    name: list[int]
    delegate: str
    account_num: int
    being_liquidated: int
    in_health_region: int
    bump: int
    padding: list[int]
    net_deposits: int
    perp_spot_transfers: int
    health_region_begin_init_health: int
    frozen_until: int
    buyback_fees_accrued_current: int
    buyback_fees_accrued_previous: int
    buyback_fees_expiry_timestamp: int
    reserved: list[int]
    header_version: int
    padding3: list[int]
    padding4: int
    tokens: list[types.token_position.TokenPositionJSON]
    padding5: int
    serum3: list[types.serum3_orders.Serum3OrdersJSON]
    padding6: int
    perps: list[types.perp_position.PerpPositionJSON]
    padding7: int
    perp_open_orders: list[types.perp_open_order.PerpOpenOrderJSON]


@dataclass
class MangoAccount:
    discriminator: typing.ClassVar = b"\xf3\xe4\xf7\x03\xa94\xaf\x1f"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "owner" / BorshPubkey,
        "name" / borsh.U8[32],
        "delegate" / BorshPubkey,
        "account_num" / borsh.U32,
        "being_liquidated" / borsh.U8,
        "in_health_region" / borsh.U8,
        "bump" / borsh.U8,
        "padding" / borsh.U8[1],
        "net_deposits" / borsh.I64,
        "perp_spot_transfers" / borsh.I64,
        "health_region_begin_init_health" / borsh.I64,
        "frozen_until" / borsh.U64,
        "buyback_fees_accrued_current" / borsh.U64,
        "buyback_fees_accrued_previous" / borsh.U64,
        "buyback_fees_expiry_timestamp" / borsh.U64,
        "reserved" / borsh.U8[208],
        "header_version" / borsh.U8,
        "padding3" / borsh.U8[7],
        "padding4" / borsh.U32,
        "tokens"
        / borsh.Vec(typing.cast(Construct, types.token_position.TokenPosition.layout)),
        "padding5" / borsh.U32,
        "serum3"
        / borsh.Vec(typing.cast(Construct, types.serum3_orders.Serum3Orders.layout)),
        "padding6" / borsh.U32,
        "perps"
        / borsh.Vec(typing.cast(Construct, types.perp_position.PerpPosition.layout)),
        "padding7" / borsh.U32,
        "perp_open_orders"
        / borsh.Vec(typing.cast(Construct, types.perp_open_order.PerpOpenOrder.layout)),
    )
    group: PublicKey
    owner: PublicKey
    name: list[int]
    delegate: PublicKey
    account_num: int
    being_liquidated: int
    in_health_region: int
    bump: int
    padding: list[int]
    net_deposits: int
    perp_spot_transfers: int
    health_region_begin_init_health: int
    frozen_until: int
    buyback_fees_accrued_current: int
    buyback_fees_accrued_previous: int
    buyback_fees_expiry_timestamp: int
    reserved: list[int]
    header_version: int
    padding3: list[int]
    padding4: int
    tokens: list[types.token_position.TokenPosition]
    padding5: int
    serum3: list[types.serum3_orders.Serum3Orders]
    padding6: int
    perps: list[types.perp_position.PerpPosition]
    padding7: int
    perp_open_orders: list[types.perp_open_order.PerpOpenOrder]
    public_key: PublicKey = None

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["MangoAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        mango_account = cls.decode(bytes_data, address)
        return mango_account

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["MangoAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MangoAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            mango_account = cls.decode(info.account.data, info.pubkey)
            res.append(mango_account)
        return res

    @classmethod
    def decode(cls, data: bytes, public_key: PublicKey = None) -> "MangoAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = MangoAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            group=dec.group,
            owner=dec.owner,
            name=dec.name,
            delegate=dec.delegate,
            account_num=dec.account_num,
            being_liquidated=dec.being_liquidated,
            in_health_region=dec.in_health_region,
            bump=dec.bump,
            padding=dec.padding,
            net_deposits=dec.net_deposits,
            perp_spot_transfers=dec.perp_spot_transfers,
            health_region_begin_init_health=dec.health_region_begin_init_health,
            frozen_until=dec.frozen_until,
            buyback_fees_accrued_current=dec.buyback_fees_accrued_current,
            buyback_fees_accrued_previous=dec.buyback_fees_accrued_previous,
            buyback_fees_expiry_timestamp=dec.buyback_fees_expiry_timestamp,
            reserved=dec.reserved,
            header_version=dec.header_version,
            padding3=dec.padding3,
            padding4=dec.padding4,
            tokens=list(
                map(
                    lambda item: types.token_position.TokenPosition.from_decoded(item),
                    dec.tokens,
                )
            ),
            padding5=dec.padding5,
            serum3=list(
                map(
                    lambda item: types.serum3_orders.Serum3Orders.from_decoded(item),
                    dec.serum3,
                )
            ),
            padding6=dec.padding6,
            perps=list(
                map(
                    lambda item: types.perp_position.PerpPosition.from_decoded(item),
                    dec.perps,
                )
            ),
            padding7=dec.padding7,
            perp_open_orders=list(
                map(
                    lambda item: types.perp_open_order.PerpOpenOrder.from_decoded(item),
                    dec.perp_open_orders,
                )
            ),
            public_key=public_key
        )

    def to_json(self) -> MangoAccountJSON:
        return {
            "group": str(self.group),
            "owner": str(self.owner),
            "name": self.name,
            "delegate": str(self.delegate),
            "account_num": self.account_num,
            "being_liquidated": self.being_liquidated,
            "in_health_region": self.in_health_region,
            "bump": self.bump,
            "padding": self.padding,
            "net_deposits": self.net_deposits,
            "perp_spot_transfers": self.perp_spot_transfers,
            "health_region_begin_init_health": self.health_region_begin_init_health,
            "frozen_until": self.frozen_until,
            "buyback_fees_accrued_current": self.buyback_fees_accrued_current,
            "buyback_fees_accrued_previous": self.buyback_fees_accrued_previous,
            "buyback_fees_expiry_timestamp": self.buyback_fees_expiry_timestamp,
            "reserved": self.reserved,
            "header_version": self.header_version,
            "padding3": self.padding3,
            "padding4": self.padding4,
            "tokens": list(map(lambda item: item.to_json(), self.tokens)),
            "padding5": self.padding5,
            "serum3": list(map(lambda item: item.to_json(), self.serum3)),
            "padding6": self.padding6,
            "perps": list(map(lambda item: item.to_json(), self.perps)),
            "padding7": self.padding7,
            "perp_open_orders": list(
                map(lambda item: item.to_json(), self.perp_open_orders)
            ),
        }

    @classmethod
    def from_json(cls, obj: MangoAccountJSON) -> "MangoAccount":
        return cls(
            group=PublicKey(obj["group"]),
            owner=PublicKey(obj["owner"]),
            name=obj["name"],
            delegate=PublicKey(obj["delegate"]),
            account_num=obj["account_num"],
            being_liquidated=obj["being_liquidated"],
            in_health_region=obj["in_health_region"],
            bump=obj["bump"],
            padding=obj["padding"],
            net_deposits=obj["net_deposits"],
            perp_spot_transfers=obj["perp_spot_transfers"],
            health_region_begin_init_health=obj["health_region_begin_init_health"],
            frozen_until=obj["frozen_until"],
            buyback_fees_accrued_current=obj["buyback_fees_accrued_current"],
            buyback_fees_accrued_previous=obj["buyback_fees_accrued_previous"],
            buyback_fees_expiry_timestamp=obj["buyback_fees_expiry_timestamp"],
            reserved=obj["reserved"],
            header_version=obj["header_version"],
            padding3=obj["padding3"],
            padding4=obj["padding4"],
            tokens=list(
                map(
                    lambda item: types.token_position.TokenPosition.from_json(item),
                    obj["tokens"],
                )
            ),
            padding5=obj["padding5"],
            serum3=list(
                map(
                    lambda item: types.serum3_orders.Serum3Orders.from_json(item),
                    obj["serum3"],
                )
            ),
            padding6=obj["padding6"],
            perps=list(
                map(
                    lambda item: types.perp_position.PerpPosition.from_json(item),
                    obj["perps"],
                )
            ),
            padding7=obj["padding7"],
            perp_open_orders=list(
                map(
                    lambda item: types.perp_open_order.PerpOpenOrder.from_json(item),
                    obj["perp_open_orders"],
                )
            ),
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class InterestRateParamsJSON(typing.TypedDict):
    util0: float
    rate0: float
    util1: float
    rate1: float
    max_rate: float
    adjustment_factor: float


@dataclass
class InterestRateParams:
    layout: typing.ClassVar = borsh.CStruct(
        "util0" / borsh.F32,
        "rate0" / borsh.F32,
        "util1" / borsh.F32,
        "rate1" / borsh.F32,
        "max_rate" / borsh.F32,
        "adjustment_factor" / borsh.F32,
    )
    util0: float
    rate0: float
    util1: float
    rate1: float
    max_rate: float
    adjustment_factor: float

    @classmethod
    def from_decoded(cls, obj: Container) -> "InterestRateParams":
        return cls(
            util0=obj.util0,
            rate0=obj.rate0,
            util1=obj.util1,
            rate1=obj.rate1,
            max_rate=obj.max_rate,
            adjustment_factor=obj.adjustment_factor,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "util0": self.util0,
            "rate0": self.rate0,
            "util1": self.util1,
            "rate1": self.rate1,
            "max_rate": self.max_rate,
            "adjustment_factor": self.adjustment_factor,
        }

    def to_json(self) -> InterestRateParamsJSON:
        return {
            "util0": self.util0,
            "rate0": self.rate0,
            "util1": self.util1,
            "rate1": self.rate1,
            "max_rate": self.max_rate,
            "adjustment_factor": self.adjustment_factor,
        }

    @classmethod
    def from_json(cls, obj: InterestRateParamsJSON) -> "InterestRateParams":
        return cls(
            util0=obj["util0"],
            rate0=obj["rate0"],
            util1=obj["util1"],
            rate1=obj["rate1"],
            max_rate=obj["max_rate"],
            adjustment_factor=obj["adjustment_factor"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class PerpEditMarketArgs(typing.TypedDict):
    oracle_opt: typing.Optional[PublicKey]
    oracle_config_opt: typing.Optional[types.oracle_config_params.OracleConfigParams]
    base_decimals_opt: typing.Optional[int]
    maint_base_asset_weight_opt: typing.Optional[float]
    init_base_asset_weight_opt: typing.Optional[float]
    maint_base_liab_weight_opt: typing.Optional[float]
    init_base_liab_weight_opt: typing.Optional[float]
    maint_overall_asset_weight_opt: typing.Optional[float]
    init_overall_asset_weight_opt: typing.Optional[float]
    base_liquidation_fee_opt: typing.Optional[float]
    maker_fee_opt: typing.Optional[float]
    taker_fee_opt: typing.Optional[float]
    min_funding_opt: typing.Optional[float]
    max_funding_opt: typing.Optional[float]
    impact_quantity_opt: typing.Optional[int]
    group_insurance_fund_opt: typing.Optional[bool]
    fee_penalty_opt: typing.Optional[float]
    settle_fee_flat_opt: typing.Optional[float]
    settle_fee_amount_threshold_opt: typing.Optional[float]
    settle_fee_fraction_low_health_opt: typing.Optional[float]
    stable_price_delay_interval_seconds_opt: typing.Optional[int]
    stable_price_delay_growth_limit_opt: typing.Optional[float]
    stable_price_growth_limit_opt: typing.Optional[float]
    settle_pnl_limit_factor_opt: typing.Optional[float]
    settle_pnl_limit_window_size_ts_opt: typing.Optional[int]
    reduce_only_opt: typing.Optional[bool]
    reset_stable_price: bool
    positive_pnl_liquidation_fee_opt: typing.Optional[float]
    name_opt: typing.Optional[str]


layout = borsh.CStruct(
    "oracle_opt" / borsh.Option(BorshPubkey),
    "oracle_config_opt"
    / borsh.Option(types.oracle_config_params.OracleConfigParams.layout),
    "base_decimals_opt" / borsh.Option(borsh.U8),
    "maint_base_asset_weight_opt" / borsh.Option(borsh.F32),
    "init_base_asset_weight_opt" / borsh.Option(borsh.F32),
    "maint_base_liab_weight_opt" / borsh.Option(borsh.F32),
    "init_base_liab_weight_opt" / borsh.Option(borsh.F32),
    "maint_overall_asset_weight_opt" / borsh.Option(borsh.F32),
    "init_overall_asset_weight_opt" / borsh.Option(borsh.F32),
    "base_liquidation_fee_opt" / borsh.Option(borsh.F32),
    "maker_fee_opt" / borsh.Option(borsh.F32),
    "taker_fee_opt" / borsh.Option(borsh.F32),
    "min_funding_opt" / borsh.Option(borsh.F32),
    "max_funding_opt" / borsh.Option(borsh.F32),
    "impact_quantity_opt" / borsh.Option(borsh.I64),
    "group_insurance_fund_opt" / borsh.Option(borsh.Bool),
    "fee_penalty_opt" / borsh.Option(borsh.F32),
    "settle_fee_flat_opt" / borsh.Option(borsh.F32),
    "settle_fee_amount_threshold_opt" / borsh.Option(borsh.F32),
    "settle_fee_fraction_low_health_opt" / borsh.Option(borsh.F32),
    "stable_price_delay_interval_seconds_opt" / borsh.Option(borsh.U32),
    "stable_price_delay_growth_limit_opt" / borsh.Option(borsh.F32),
    "stable_price_growth_limit_opt" / borsh.Option(borsh.F32),
    "settle_pnl_limit_factor_opt" / borsh.Option(borsh.F32),
    "settle_pnl_limit_window_size_ts_opt" / borsh.Option(borsh.U64),
    "reduce_only_opt" / borsh.Option(borsh.Bool),
    "reset_stable_price" / borsh.Bool,
    "positive_pnl_liquidation_fee_opt" / borsh.Option(borsh.F32),
    "name_opt" / borsh.Option(borsh.String),
)


class PerpEditMarketAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    perp_market: PublicKey
    oracle: PublicKey


def perp_edit_market(
    args: PerpEditMarketArgs,
    accounts: PerpEditMarketAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"|r\xa0\xe7E\xdfLQ"
    encoded_args = layout.build(
        {
            "oracle_opt": args["oracle_opt"],
            "oracle_config_opt": (
                None
                if args["oracle_config_opt"] is None
                else args["oracle_config_opt"].to_encodable()
            ),
            "base_decimals_opt": args["base_decimals_opt"],
            "maint_base_asset_weight_opt": args["maint_base_asset_weight_opt"],
            "init_base_asset_weight_opt": args["init_base_asset_weight_opt"],
            "maint_base_liab_weight_opt": args["maint_base_liab_weight_opt"],
            "init_base_liab_weight_opt": args["init_base_liab_weight_opt"],
            "maint_overall_asset_weight_opt": args["maint_overall_asset_weight_opt"],
            "init_overall_asset_weight_opt": args["init_overall_asset_weight_opt"],
            "base_liquidation_fee_opt": args["base_liquidation_fee_opt"],
            "maker_fee_opt": args["maker_fee_opt"],
            "taker_fee_opt": args["taker_fee_opt"],
            "min_funding_opt": args["min_funding_opt"],
            "max_funding_opt": args["max_funding_opt"],
            "impact_quantity_opt": args["impact_quantity_opt"],
            "group_insurance_fund_opt": args["group_insurance_fund_opt"],
            "fee_penalty_opt": args["fee_penalty_opt"],
            "settle_fee_flat_opt": args["settle_fee_flat_opt"],
            "settle_fee_amount_threshold_opt": args["settle_fee_amount_threshold_opt"],
            "settle_fee_fraction_low_health_opt": args[
                "settle_fee_fraction_low_health_opt"
            ],
            "stable_price_delay_interval_seconds_opt": args[
                "stable_price_delay_interval_seconds_opt"
            ],
            "stable_price_delay_growth_limit_opt": args[
                "stable_price_delay_growth_limit_opt"
            ],
            "stable_price_growth_limit_opt": args["stable_price_growth_limit_opt"],
            "settle_pnl_limit_factor_opt": args["settle_pnl_limit_factor_opt"],
            "settle_pnl_limit_window_size_ts_opt": args[
                "settle_pnl_limit_window_size_ts_opt"
            ],
            "reduce_only_opt": args["reduce_only_opt"],
            "reset_stable_price": args["reset_stable_price"],
            "positive_pnl_liquidation_fee_opt": args[
                "positive_pnl_liquidation_fee_opt"
            ],
            "name_opt": args["name_opt"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class PythJSON(typing.TypedDict):
    kind: typing.Literal["Pyth"]


class StubJSON(typing.TypedDict):
    kind: typing.Literal["Stub"]


class SwitchboardV1JSON(typing.TypedDict):
    kind: typing.Literal["SwitchboardV1"]


class SwitchboardV2JSON(typing.TypedDict):
    kind: typing.Literal["SwitchboardV2"]


@dataclass
class Pyth:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Pyth"

    @classmethod
    def to_json(cls) -> PythJSON:
        return PythJSON(
            kind="Pyth",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Pyth": {},
        }


@dataclass
class Stub:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Stub"

    @classmethod
    def to_json(cls) -> StubJSON:
        return StubJSON(
            kind="Stub",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Stub": {},
        }


@dataclass
class SwitchboardV1:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "SwitchboardV1"

    @classmethod
    def to_json(cls) -> SwitchboardV1JSON:
        return SwitchboardV1JSON(
            kind="SwitchboardV1",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SwitchboardV1": {},
        }


@dataclass
class SwitchboardV2:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "SwitchboardV2"

    @classmethod
    def to_json(cls) -> SwitchboardV2JSON:
        return SwitchboardV2JSON(
            kind="SwitchboardV2",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SwitchboardV2": {},
        }


OracleTypeKind = typing.Union[Pyth, Stub, SwitchboardV1, SwitchboardV2]
OracleTypeJSON = typing.Union[PythJSON, StubJSON, SwitchboardV1JSON, SwitchboardV2JSON]


def from_decoded(obj: dict) -> OracleTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Pyth" in obj:
        return Pyth()
    if "Stub" in obj:
        return Stub()
    if "SwitchboardV1" in obj:
        return SwitchboardV1()
    if "SwitchboardV2" in obj:
        return SwitchboardV2()
    raise ValueError("Invalid enum object")


def from_json(obj: OracleTypeJSON) -> OracleTypeKind:
    if obj["kind"] == "Pyth":
        return Pyth()
    if obj["kind"] == "Stub":
        return Stub()
    if obj["kind"] == "SwitchboardV1":
        return SwitchboardV1()
    if obj["kind"] == "SwitchboardV2":
        return SwitchboardV2()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Pyth" / borsh.CStruct(),
    "Stub" / borsh.CStruct(),
    "SwitchboardV1" / borsh.CStruct(),
    "SwitchboardV2" / borsh.CStruct(),
)



import typing
from anchorpy.error import ProgramError


class SomeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "")

    code = 6000
    name = "SomeError"
    msg = ""


class NotImplementedError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "")

    code = 6001
    name = "NotImplementedError"
    msg = ""


class MathError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "checked math error")

    code = 6002
    name = "MathError"
    msg = "checked math error"


class UnexpectedOracle(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "")

    code = 6003
    name = "UnexpectedOracle"
    msg = ""


class UnknownOracleType(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "oracle type cannot be determined")

    code = 6004
    name = "UnknownOracleType"
    msg = "oracle type cannot be determined"


class InvalidFlashLoanTargetCpiProgram(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "")

    code = 6005
    name = "InvalidFlashLoanTargetCpiProgram"
    msg = ""


class HealthMustBePositive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "health must be positive")

    code = 6006
    name = "HealthMustBePositive"
    msg = "health must be positive"


class HealthMustBePositiveOrIncrease(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "health must be positive or increase")

    code = 6007
    name = "HealthMustBePositiveOrIncrease"
    msg = "health must be positive or increase"


class HealthMustBeNegative(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "health must be negative")

    code = 6008
    name = "HealthMustBeNegative"
    msg = "health must be negative"


class IsBankrupt(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "the account is bankrupt")

    code = 6009
    name = "IsBankrupt"
    msg = "the account is bankrupt"


class IsNotBankrupt(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "the account is not bankrupt")

    code = 6010
    name = "IsNotBankrupt"
    msg = "the account is not bankrupt"


class NoFreeTokenPositionIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "no free token position index")

    code = 6011
    name = "NoFreeTokenPositionIndex"
    msg = "no free token position index"


class NoFreeSerum3OpenOrdersIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "no free serum3 open orders index")

    code = 6012
    name = "NoFreeSerum3OpenOrdersIndex"
    msg = "no free serum3 open orders index"


class NoFreePerpPositionIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "no free perp position index")

    code = 6013
    name = "NoFreePerpPositionIndex"
    msg = "no free perp position index"


class Serum3OpenOrdersExistAlready(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "serum3 open orders exist already")

    code = 6014
    name = "Serum3OpenOrdersExistAlready"
    msg = "serum3 open orders exist already"


class InsufficentBankVaultFunds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "bank vault has insufficent funds")

    code = 6015
    name = "InsufficentBankVaultFunds"
    msg = "bank vault has insufficent funds"


class BeingLiquidated(ProgramError):
    def __init__(self) -> None:
        super().__init__(6016, "account is currently being liquidated")

    code = 6016
    name = "BeingLiquidated"
    msg = "account is currently being liquidated"


class InvalidBank(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "invalid bank")

    code = 6017
    name = "InvalidBank"
    msg = "invalid bank"


class ProfitabilityMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "account profitability is mismatched")

    code = 6018
    name = "ProfitabilityMismatch"
    msg = "account profitability is mismatched"


class CannotSettleWithSelf(ProgramError):
    def __init__(self) -> None:
        super().__init__(6019, "cannot settle with self")

    code = 6019
    name = "CannotSettleWithSelf"
    msg = "cannot settle with self"


class PerpPositionDoesNotExist(ProgramError):
    def __init__(self) -> None:
        super().__init__(6020, "perp position does not exist")

    code = 6020
    name = "PerpPositionDoesNotExist"
    msg = "perp position does not exist"


class MaxSettleAmountMustBeGreaterThanZero(ProgramError):
    def __init__(self) -> None:
        super().__init__(6021, "max settle amount must be greater than zero")

    code = 6021
    name = "MaxSettleAmountMustBeGreaterThanZero"
    msg = "max settle amount must be greater than zero"


class HasOpenPerpOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6022, "the perp position has open orders or unprocessed fill events"
        )

    code = 6022
    name = "HasOpenPerpOrders"
    msg = "the perp position has open orders or unprocessed fill events"


class OracleConfidence(ProgramError):
    def __init__(self) -> None:
        super().__init__(6023, "an oracle does not reach the confidence threshold")

    code = 6023
    name = "OracleConfidence"
    msg = "an oracle does not reach the confidence threshold"


class OracleStale(ProgramError):
    def __init__(self) -> None:
        super().__init__(6024, "an oracle is stale")

    code = 6024
    name = "OracleStale"
    msg = "an oracle is stale"


class SettlementAmountMustBePositive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6025, "settlement amount must always be positive")

    code = 6025
    name = "SettlementAmountMustBePositive"
    msg = "settlement amount must always be positive"


class BankBorrowLimitReached(ProgramError):
    def __init__(self) -> None:
        super().__init__(6026, "bank utilization has reached limit")

    code = 6026
    name = "BankBorrowLimitReached"
    msg = "bank utilization has reached limit"


class BankNetBorrowsLimitReached(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6027,
            "bank net borrows has reached limit - this is an intermittent error - the limit will reset regularly",
        )

    code = 6027
    name = "BankNetBorrowsLimitReached"
    msg = "bank net borrows has reached limit - this is an intermittent error - the limit will reset regularly"


class TokenPositionDoesNotExist(ProgramError):
    def __init__(self) -> None:
        super().__init__(6028, "token position does not exist")

    code = 6028
    name = "TokenPositionDoesNotExist"
    msg = "token position does not exist"


class DepositsIntoLiquidatingMustRecover(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6029,
            "token deposits into accounts that are being liquidated must bring their health above the init threshold",
        )

    code = 6029
    name = "DepositsIntoLiquidatingMustRecover"
    msg = "token deposits into accounts that are being liquidated must bring their health above the init threshold"


class TokenInReduceOnlyMode(ProgramError):
    def __init__(self) -> None:
        super().__init__(6030, "token is in reduce only mode")

    code = 6030
    name = "TokenInReduceOnlyMode"
    msg = "token is in reduce only mode"


class MarketInReduceOnlyMode(ProgramError):
    def __init__(self) -> None:
        super().__init__(6031, "market is in reduce only mode")

    code = 6031
    name = "MarketInReduceOnlyMode"
    msg = "market is in reduce only mode"


class GroupIsHalted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6032, "group is halted")

    code = 6032
    name = "GroupIsHalted"
    msg = "group is halted"


class PerpHasBaseLots(ProgramError):
    def __init__(self) -> None:
        super().__init__(6033, "the perp position has non-zero base lots")

    code = 6033
    name = "PerpHasBaseLots"
    msg = "the perp position has non-zero base lots"


class HasOpenOrUnsettledSerum3Orders(ProgramError):
    def __init__(self) -> None:
        super().__init__(6034, "there are open or unsettled serum3 orders")

    code = 6034
    name = "HasOpenOrUnsettledSerum3Orders"
    msg = "there are open or unsettled serum3 orders"


class HasLiquidatableTokenPosition(ProgramError):
    def __init__(self) -> None:
        super().__init__(6035, "has liquidatable token position")

    code = 6035
    name = "HasLiquidatableTokenPosition"
    msg = "has liquidatable token position"


class HasLiquidatablePerpBasePosition(ProgramError):
    def __init__(self) -> None:
        super().__init__(6036, "has liquidatable perp base position")

    code = 6036
    name = "HasLiquidatablePerpBasePosition"
    msg = "has liquidatable perp base position"


class HasLiquidatablePositivePerpPnl(ProgramError):
    def __init__(self) -> None:
        super().__init__(6037, "has liquidatable positive perp pnl")

    code = 6037
    name = "HasLiquidatablePositivePerpPnl"
    msg = "has liquidatable positive perp pnl"


class AccountIsFrozen(ProgramError):
    def __init__(self) -> None:
        super().__init__(6038, "account is frozen")

    code = 6038
    name = "AccountIsFrozen"
    msg = "account is frozen"


class InitAssetWeightCantBeNegative(ProgramError):
    def __init__(self) -> None:
        super().__init__(6039, "Init Asset Weight can't be negative")

    code = 6039
    name = "InitAssetWeightCantBeNegative"
    msg = "Init Asset Weight can't be negative"


class HasOpenPerpTakerFills(ProgramError):
    def __init__(self) -> None:
        super().__init__(6040, "has open perp taker fills")

    code = 6040
    name = "HasOpenPerpTakerFills"
    msg = "has open perp taker fills"


class DepositLimit(ProgramError):
    def __init__(self) -> None:
        super().__init__(6041, "deposit crosses the current group deposit limit")

    code = 6041
    name = "DepositLimit"
    msg = "deposit crosses the current group deposit limit"


class IxIsDisabled(ProgramError):
    def __init__(self) -> None:
        super().__init__(6042, "instruction is disabled")

    code = 6042
    name = "IxIsDisabled"
    msg = "instruction is disabled"


class NoLiquidatablePerpBasePosition(ProgramError):
    def __init__(self) -> None:
        super().__init__(6043, "no liquidatable perp base position")

    code = 6043
    name = "NoLiquidatablePerpBasePosition"
    msg = "no liquidatable perp base position"


class PerpOrderIdNotFound(ProgramError):
    def __init__(self) -> None:
        super().__init__(6044, "perp order id not found on the orderbook")

    code = 6044
    name = "PerpOrderIdNotFound"
    msg = "perp order id not found on the orderbook"


CustomError = typing.Union[
    SomeError,
    NotImplementedError,
    MathError,
    UnexpectedOracle,
    UnknownOracleType,
    InvalidFlashLoanTargetCpiProgram,
    HealthMustBePositive,
    HealthMustBePositiveOrIncrease,
    HealthMustBeNegative,
    IsBankrupt,
    IsNotBankrupt,
    NoFreeTokenPositionIndex,
    NoFreeSerum3OpenOrdersIndex,
    NoFreePerpPositionIndex,
    Serum3OpenOrdersExistAlready,
    InsufficentBankVaultFunds,
    BeingLiquidated,
    InvalidBank,
    ProfitabilityMismatch,
    CannotSettleWithSelf,
    PerpPositionDoesNotExist,
    MaxSettleAmountMustBeGreaterThanZero,
    HasOpenPerpOrders,
    OracleConfidence,
    OracleStale,
    SettlementAmountMustBePositive,
    BankBorrowLimitReached,
    BankNetBorrowsLimitReached,
    TokenPositionDoesNotExist,
    DepositsIntoLiquidatingMustRecover,
    TokenInReduceOnlyMode,
    MarketInReduceOnlyMode,
    GroupIsHalted,
    PerpHasBaseLots,
    HasOpenOrUnsettledSerum3Orders,
    HasLiquidatableTokenPosition,
    HasLiquidatablePerpBasePosition,
    HasLiquidatablePositivePerpPnl,
    AccountIsFrozen,
    InitAssetWeightCantBeNegative,
    HasOpenPerpTakerFills,
    DepositLimit,
    IxIsDisabled,
    NoLiquidatablePerpBasePosition,
    PerpOrderIdNotFound,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: SomeError(),
    6001: NotImplementedError(),
    6002: MathError(),
    6003: UnexpectedOracle(),
    6004: UnknownOracleType(),
    6005: InvalidFlashLoanTargetCpiProgram(),
    6006: HealthMustBePositive(),
    6007: HealthMustBePositiveOrIncrease(),
    6008: HealthMustBeNegative(),
    6009: IsBankrupt(),
    6010: IsNotBankrupt(),
    6011: NoFreeTokenPositionIndex(),
    6012: NoFreeSerum3OpenOrdersIndex(),
    6013: NoFreePerpPositionIndex(),
    6014: Serum3OpenOrdersExistAlready(),
    6015: InsufficentBankVaultFunds(),
    6016: BeingLiquidated(),
    6017: InvalidBank(),
    6018: ProfitabilityMismatch(),
    6019: CannotSettleWithSelf(),
    6020: PerpPositionDoesNotExist(),
    6021: MaxSettleAmountMustBeGreaterThanZero(),
    6022: HasOpenPerpOrders(),
    6023: OracleConfidence(),
    6024: OracleStale(),
    6025: SettlementAmountMustBePositive(),
    6026: BankBorrowLimitReached(),
    6027: BankNetBorrowsLimitReached(),
    6028: TokenPositionDoesNotExist(),
    6029: DepositsIntoLiquidatingMustRecover(),
    6030: TokenInReduceOnlyMode(),
    6031: MarketInReduceOnlyMode(),
    6032: GroupIsHalted(),
    6033: PerpHasBaseLots(),
    6034: HasOpenOrUnsettledSerum3Orders(),
    6035: HasLiquidatableTokenPosition(),
    6036: HasLiquidatablePerpBasePosition(),
    6037: HasLiquidatablePositivePerpPnl(),
    6038: AccountIsFrozen(),
    6039: InitAssetWeightCantBeNegative(),
    6040: HasOpenPerpTakerFills(),
    6041: DepositLimit(),
    6042: IxIsDisabled(),
    6043: NoLiquidatablePerpBasePosition(),
    6044: PerpOrderIdNotFound(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solana.publickey import PublicKey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class Serum3OrdersJSON(typing.TypedDict):
    open_orders: str
    base_borrows_without_fee: int
    quote_borrows_without_fee: int
    market_index: int
    base_token_index: int
    quote_token_index: int
    padding: list[int]
    reserved: list[int]


@dataclass
class Serum3Orders:
    layout: typing.ClassVar = borsh.CStruct(
        "open_orders" / BorshPubkey,
        "base_borrows_without_fee" / borsh.U64,
        "quote_borrows_without_fee" / borsh.U64,
        "market_index" / borsh.U16,
        "base_token_index" / borsh.U16,
        "quote_token_index" / borsh.U16,
        "padding" / borsh.U8[2],
        "reserved" / borsh.U8[64],
    )
    open_orders: PublicKey
    base_borrows_without_fee: int
    quote_borrows_without_fee: int
    market_index: int
    base_token_index: int
    quote_token_index: int
    padding: list[int]
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "Serum3Orders":
        return cls(
            open_orders=obj.open_orders,
            base_borrows_without_fee=obj.base_borrows_without_fee,
            quote_borrows_without_fee=obj.quote_borrows_without_fee,
            market_index=obj.market_index,
            base_token_index=obj.base_token_index,
            quote_token_index=obj.quote_token_index,
            padding=obj.padding,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "open_orders": self.open_orders,
            "base_borrows_without_fee": self.base_borrows_without_fee,
            "quote_borrows_without_fee": self.quote_borrows_without_fee,
            "market_index": self.market_index,
            "base_token_index": self.base_token_index,
            "quote_token_index": self.quote_token_index,
            "padding": self.padding,
            "reserved": self.reserved,
        }

    def to_json(self) -> Serum3OrdersJSON:
        return {
            "open_orders": str(self.open_orders),
            "base_borrows_without_fee": self.base_borrows_without_fee,
            "quote_borrows_without_fee": self.quote_borrows_without_fee,
            "market_index": self.market_index,
            "base_token_index": self.base_token_index,
            "quote_token_index": self.quote_token_index,
            "padding": self.padding,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: Serum3OrdersJSON) -> "Serum3Orders":
        return cls(
            open_orders=PublicKey(obj["open_orders"]),
            base_borrows_without_fee=obj["base_borrows_without_fee"],
            quote_borrows_without_fee=obj["quote_borrows_without_fee"],
            market_index=obj["market_index"],
            base_token_index=obj["base_token_index"],
            quote_token_index=obj["quote_token_index"],
            padding=obj["padding"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TokenPositionJSON(typing.TypedDict):
    indexed_position: i80f48.I80F48JSON
    token_index: int
    in_use_count: int
    padding: list[int]
    previous_index: i80f48.I80F48JSON
    cumulative_deposit_interest: float
    cumulative_borrow_interest: float
    reserved: list[int]


@dataclass
class TokenPosition:
    layout: typing.ClassVar = borsh.CStruct(
        "indexed_position" / i80f48.I80F48.layout,
        "token_index" / borsh.U16,
        "in_use_count" / borsh.U8,
        "padding" / borsh.U8[5],
        "previous_index" / i80f48.I80F48.layout,
        "cumulative_deposit_interest" / borsh.F64,
        "cumulative_borrow_interest" / borsh.F64,
        "reserved" / borsh.U8[128],
    )
    indexed_position: i80f48.I80F48
    token_index: int
    in_use_count: int
    padding: list[int]
    previous_index: i80f48.I80F48
    cumulative_deposit_interest: float
    cumulative_borrow_interest: float
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "TokenPosition":
        return cls(
            indexed_position=i80f48.I80F48.from_decoded(obj.indexed_position),
            token_index=obj.token_index,
            in_use_count=obj.in_use_count,
            padding=obj.padding,
            previous_index=i80f48.I80F48.from_decoded(obj.previous_index),
            cumulative_deposit_interest=obj.cumulative_deposit_interest,
            cumulative_borrow_interest=obj.cumulative_borrow_interest,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "indexed_position": self.indexed_position.to_encodable(),
            "token_index": self.token_index,
            "in_use_count": self.in_use_count,
            "padding": self.padding,
            "previous_index": self.previous_index.to_encodable(),
            "cumulative_deposit_interest": self.cumulative_deposit_interest,
            "cumulative_borrow_interest": self.cumulative_borrow_interest,
            "reserved": self.reserved,
        }

    def to_json(self) -> TokenPositionJSON:
        return {
            "indexed_position": self.indexed_position.to_json(),
            "token_index": self.token_index,
            "in_use_count": self.in_use_count,
            "padding": self.padding,
            "previous_index": self.previous_index.to_json(),
            "cumulative_deposit_interest": self.cumulative_deposit_interest,
            "cumulative_borrow_interest": self.cumulative_borrow_interest,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: TokenPositionJSON) -> "TokenPosition":
        return cls(
            indexed_position=i80f48.I80F48.from_json(obj["indexed_position"]),
            token_index=obj["token_index"],
            in_use_count=obj["in_use_count"],
            padding=obj["padding"],
            previous_index=i80f48.I80F48.from_json(obj["previous_index"]),
            cumulative_deposit_interest=obj["cumulative_deposit_interest"],
            cumulative_borrow_interest=obj["cumulative_borrow_interest"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PerpPositionJSON(typing.TypedDict):
    market_index: int
    padding: list[int]
    settle_pnl_limit_window: int
    settle_pnl_limit_settled_in_current_window_native: int
    base_position_lots: int
    quote_position_native: i80f48.I80F48JSON
    quote_running_native: int
    long_settled_funding: i80f48.I80F48JSON
    short_settled_funding: i80f48.I80F48JSON
    bids_base_lots: int
    asks_base_lots: int
    taker_base_lots: int
    taker_quote_lots: int
    cumulative_long_funding: float
    cumulative_short_funding: float
    maker_volume: int
    taker_volume: int
    perp_spot_transfers: int
    avg_entry_price_per_base_lot: float
    realized_trade_pnl_native: i80f48.I80F48JSON
    realized_other_pnl_native: i80f48.I80F48JSON
    settle_pnl_limit_realized_trade: int
    realized_pnl_for_position_native: i80f48.I80F48JSON
    reserved: list[int]


@dataclass
class PerpPosition:
    layout: typing.ClassVar = borsh.CStruct(
        "market_index" / borsh.U16,
        "padding" / borsh.U8[2],
        "settle_pnl_limit_window" / borsh.U32,
        "settle_pnl_limit_settled_in_current_window_native" / borsh.I64,
        "base_position_lots" / borsh.I64,
        "quote_position_native" / i80f48.I80F48.layout,
        "quote_running_native" / borsh.I64,
        "long_settled_funding" / i80f48.I80F48.layout,
        "short_settled_funding" / i80f48.I80F48.layout,
        "bids_base_lots" / borsh.I64,
        "asks_base_lots" / borsh.I64,
        "taker_base_lots" / borsh.I64,
        "taker_quote_lots" / borsh.I64,
        "cumulative_long_funding" / borsh.F64,
        "cumulative_short_funding" / borsh.F64,
        "maker_volume" / borsh.U64,
        "taker_volume" / borsh.U64,
        "perp_spot_transfers" / borsh.I64,
        "avg_entry_price_per_base_lot" / borsh.F64,
        "realized_trade_pnl_native" / i80f48.I80F48.layout,
        "realized_other_pnl_native" / i80f48.I80F48.layout,
        "settle_pnl_limit_realized_trade" / borsh.I64,
        "realized_pnl_for_position_native" / i80f48.I80F48.layout,
        "reserved" / borsh.U8[88],
    )
    market_index: int
    padding: list[int]
    settle_pnl_limit_window: int
    settle_pnl_limit_settled_in_current_window_native: int
    base_position_lots: int
    quote_position_native: i80f48.I80F48
    quote_running_native: int
    long_settled_funding: i80f48.I80F48
    short_settled_funding: i80f48.I80F48
    bids_base_lots: int
    asks_base_lots: int
    taker_base_lots: int
    taker_quote_lots: int
    cumulative_long_funding: float
    cumulative_short_funding: float
    maker_volume: int
    taker_volume: int
    perp_spot_transfers: int
    avg_entry_price_per_base_lot: float
    realized_trade_pnl_native: i80f48.I80F48
    realized_other_pnl_native: i80f48.I80F48
    settle_pnl_limit_realized_trade: int
    realized_pnl_for_position_native: i80f48.I80F48
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "PerpPosition":
        return cls(
            market_index=obj.market_index,
            padding=obj.padding,
            settle_pnl_limit_window=obj.settle_pnl_limit_window,
            settle_pnl_limit_settled_in_current_window_native=obj.settle_pnl_limit_settled_in_current_window_native,
            base_position_lots=obj.base_position_lots,
            quote_position_native=i80f48.I80F48.from_decoded(obj.quote_position_native),
            quote_running_native=obj.quote_running_native,
            long_settled_funding=i80f48.I80F48.from_decoded(obj.long_settled_funding),
            short_settled_funding=i80f48.I80F48.from_decoded(obj.short_settled_funding),
            bids_base_lots=obj.bids_base_lots,
            asks_base_lots=obj.asks_base_lots,
            taker_base_lots=obj.taker_base_lots,
            taker_quote_lots=obj.taker_quote_lots,
            cumulative_long_funding=obj.cumulative_long_funding,
            cumulative_short_funding=obj.cumulative_short_funding,
            maker_volume=obj.maker_volume,
            taker_volume=obj.taker_volume,
            perp_spot_transfers=obj.perp_spot_transfers,
            avg_entry_price_per_base_lot=obj.avg_entry_price_per_base_lot,
            realized_trade_pnl_native=i80f48.I80F48.from_decoded(
                obj.realized_trade_pnl_native
            ),
            realized_other_pnl_native=i80f48.I80F48.from_decoded(
                obj.realized_other_pnl_native
            ),
            settle_pnl_limit_realized_trade=obj.settle_pnl_limit_realized_trade,
            realized_pnl_for_position_native=i80f48.I80F48.from_decoded(
                obj.realized_pnl_for_position_native
            ),
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "market_index": self.market_index,
            "padding": self.padding,
            "settle_pnl_limit_window": self.settle_pnl_limit_window,
            "settle_pnl_limit_settled_in_current_window_native": self.settle_pnl_limit_settled_in_current_window_native,
            "base_position_lots": self.base_position_lots,
            "quote_position_native": self.quote_position_native.to_encodable(),
            "quote_running_native": self.quote_running_native,
            "long_settled_funding": self.long_settled_funding.to_encodable(),
            "short_settled_funding": self.short_settled_funding.to_encodable(),
            "bids_base_lots": self.bids_base_lots,
            "asks_base_lots": self.asks_base_lots,
            "taker_base_lots": self.taker_base_lots,
            "taker_quote_lots": self.taker_quote_lots,
            "cumulative_long_funding": self.cumulative_long_funding,
            "cumulative_short_funding": self.cumulative_short_funding,
            "maker_volume": self.maker_volume,
            "taker_volume": self.taker_volume,
            "perp_spot_transfers": self.perp_spot_transfers,
            "avg_entry_price_per_base_lot": self.avg_entry_price_per_base_lot,
            "realized_trade_pnl_native": self.realized_trade_pnl_native.to_encodable(),
            "realized_other_pnl_native": self.realized_other_pnl_native.to_encodable(),
            "settle_pnl_limit_realized_trade": self.settle_pnl_limit_realized_trade,
            "realized_pnl_for_position_native": self.realized_pnl_for_position_native.to_encodable(),
            "reserved": self.reserved,
        }

    def to_json(self) -> PerpPositionJSON:
        return {
            "market_index": self.market_index,
            "padding": self.padding,
            "settle_pnl_limit_window": self.settle_pnl_limit_window,
            "settle_pnl_limit_settled_in_current_window_native": self.settle_pnl_limit_settled_in_current_window_native,
            "base_position_lots": self.base_position_lots,
            "quote_position_native": self.quote_position_native.to_json(),
            "quote_running_native": self.quote_running_native,
            "long_settled_funding": self.long_settled_funding.to_json(),
            "short_settled_funding": self.short_settled_funding.to_json(),
            "bids_base_lots": self.bids_base_lots,
            "asks_base_lots": self.asks_base_lots,
            "taker_base_lots": self.taker_base_lots,
            "taker_quote_lots": self.taker_quote_lots,
            "cumulative_long_funding": self.cumulative_long_funding,
            "cumulative_short_funding": self.cumulative_short_funding,
            "maker_volume": self.maker_volume,
            "taker_volume": self.taker_volume,
            "perp_spot_transfers": self.perp_spot_transfers,
            "avg_entry_price_per_base_lot": self.avg_entry_price_per_base_lot,
            "realized_trade_pnl_native": self.realized_trade_pnl_native.to_json(),
            "realized_other_pnl_native": self.realized_other_pnl_native.to_json(),
            "settle_pnl_limit_realized_trade": self.settle_pnl_limit_realized_trade,
            "realized_pnl_for_position_native": self.realized_pnl_for_position_native.to_json(),
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: PerpPositionJSON) -> "PerpPosition":
        return cls(
            market_index=obj["market_index"],
            padding=obj["padding"],
            settle_pnl_limit_window=obj["settle_pnl_limit_window"],
            settle_pnl_limit_settled_in_current_window_native=obj[
                "settle_pnl_limit_settled_in_current_window_native"
            ],
            base_position_lots=obj["base_position_lots"],
            quote_position_native=i80f48.I80F48.from_json(obj["quote_position_native"]),
            quote_running_native=obj["quote_running_native"],
            long_settled_funding=i80f48.I80F48.from_json(obj["long_settled_funding"]),
            short_settled_funding=i80f48.I80F48.from_json(obj["short_settled_funding"]),
            bids_base_lots=obj["bids_base_lots"],
            asks_base_lots=obj["asks_base_lots"],
            taker_base_lots=obj["taker_base_lots"],
            taker_quote_lots=obj["taker_quote_lots"],
            cumulative_long_funding=obj["cumulative_long_funding"],
            cumulative_short_funding=obj["cumulative_short_funding"],
            maker_volume=obj["maker_volume"],
            taker_volume=obj["taker_volume"],
            perp_spot_transfers=obj["perp_spot_transfers"],
            avg_entry_price_per_base_lot=obj["avg_entry_price_per_base_lot"],
            realized_trade_pnl_native=i80f48.I80F48.from_json(
                obj["realized_trade_pnl_native"]
            ),
            realized_other_pnl_native=i80f48.I80F48.from_json(
                obj["realized_other_pnl_native"]
            ),
            settle_pnl_limit_realized_trade=obj["settle_pnl_limit_realized_trade"],
            realized_pnl_for_position_native=i80f48.I80F48.from_json(
                obj["realized_pnl_for_position_native"]
            ),
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class FillJSON(typing.TypedDict):
    kind: typing.Literal["Fill"]


class OutJSON(typing.TypedDict):
    kind: typing.Literal["Out"]


class LiquidateJSON(typing.TypedDict):
    kind: typing.Literal["Liquidate"]


@dataclass
class Fill:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Fill"

    @classmethod
    def to_json(cls) -> FillJSON:
        return FillJSON(
            kind="Fill",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Fill": {},
        }


@dataclass
class Out:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Out"

    @classmethod
    def to_json(cls) -> OutJSON:
        return OutJSON(
            kind="Out",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Out": {},
        }


@dataclass
class Liquidate:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Liquidate"

    @classmethod
    def to_json(cls) -> LiquidateJSON:
        return LiquidateJSON(
            kind="Liquidate",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Liquidate": {},
        }


EventTypeKind = typing.Union[Fill, Out, Liquidate]
EventTypeJSON = typing.Union[FillJSON, OutJSON, LiquidateJSON]


def from_decoded(obj: dict) -> EventTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Fill" in obj:
        return Fill()
    if "Out" in obj:
        return Out()
    if "Liquidate" in obj:
        return Liquidate()
    raise ValueError("Invalid enum object")


def from_json(obj: EventTypeJSON) -> EventTypeKind:
    if obj["kind"] == "Fill":
        return Fill()
    if obj["kind"] == "Out":
        return Out()
    if obj["kind"] == "Liquidate":
        return Liquidate()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Fill" / borsh.CStruct(), "Out" / borsh.CStruct(), "Liquidate" / borsh.CStruct()
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


def benchmark(
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = []
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"yv;\xdcV\x8d\xa6z"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
from construct import Construct
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class FlashLoanBeginArgs(typing.TypedDict):
    loan_amounts: list[int]


layout = borsh.CStruct("loan_amounts" / borsh.Vec(typing.cast(Construct, borsh.U64)))


class FlashLoanBeginAccounts(typing.TypedDict):
    account: PublicKey
    owner: PublicKey
    instructions: PublicKey


def flash_loan_begin(
    args: FlashLoanBeginArgs,
    accounts: FlashLoanBeginAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["instructions"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"QN\xe0<\xf48Z\xef"
    encoded_args = layout.build(
        {
            "loan_amounts": args["loan_amounts"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PricesJSON(typing.TypedDict):
    oracle: i80f48.I80F48JSON
    stable: i80f48.I80F48JSON


@dataclass
class Prices:
    layout: typing.ClassVar = borsh.CStruct(
        "oracle" / i80f48.I80F48.layout, "stable" / i80f48.I80F48.layout
    )
    oracle: i80f48.I80F48
    stable: i80f48.I80F48

    @classmethod
    def from_decoded(cls, obj: Container) -> "Prices":
        return cls(
            oracle=i80f48.I80F48.from_decoded(obj.oracle),
            stable=i80f48.I80F48.from_decoded(obj.stable),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "oracle": self.oracle.to_encodable(),
            "stable": self.stable.to_encodable(),
        }

    def to_json(self) -> PricesJSON:
        return {"oracle": self.oracle.to_json(), "stable": self.stable.to_json()}

    @classmethod
    def from_json(cls, obj: PricesJSON) -> "Prices":
        return cls(
            oracle=i80f48.I80F48.from_json(obj["oracle"]),
            stable=i80f48.I80F48.from_json(obj["stable"]),
        )



from __future__ import annotations
from . import (
    prices,
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PerpInfoJSON(typing.TypedDict):
    perp_market_index: int
    maint_base_asset_weight: i80f48.I80F48JSON
    init_base_asset_weight: i80f48.I80F48JSON
    maint_base_liab_weight: i80f48.I80F48JSON
    init_base_liab_weight: i80f48.I80F48JSON
    maint_overall_asset_weight: i80f48.I80F48JSON
    init_overall_asset_weight: i80f48.I80F48JSON
    base_lot_size: int
    base_lots: int
    bids_base_lots: int
    asks_base_lots: int
    quote: i80f48.I80F48JSON
    prices: prices.PricesJSON
    has_open_orders: bool
    has_open_fills: bool


@dataclass
class PerpInfo:
    layout: typing.ClassVar = borsh.CStruct(
        "perp_market_index" / borsh.U16,
        "maint_base_asset_weight" / i80f48.I80F48.layout,
        "init_base_asset_weight" / i80f48.I80F48.layout,
        "maint_base_liab_weight" / i80f48.I80F48.layout,
        "init_base_liab_weight" / i80f48.I80F48.layout,
        "maint_overall_asset_weight" / i80f48.I80F48.layout,
        "init_overall_asset_weight" / i80f48.I80F48.layout,
        "base_lot_size" / borsh.I64,
        "base_lots" / borsh.I64,
        "bids_base_lots" / borsh.I64,
        "asks_base_lots" / borsh.I64,
        "quote" / i80f48.I80F48.layout,
        "prices" / prices.Prices.layout,
        "has_open_orders" / borsh.Bool,
        "has_open_fills" / borsh.Bool,
    )
    perp_market_index: int
    maint_base_asset_weight: i80f48.I80F48
    init_base_asset_weight: i80f48.I80F48
    maint_base_liab_weight: i80f48.I80F48
    init_base_liab_weight: i80f48.I80F48
    maint_overall_asset_weight: i80f48.I80F48
    init_overall_asset_weight: i80f48.I80F48
    base_lot_size: int
    base_lots: int
    bids_base_lots: int
    asks_base_lots: int
    quote: i80f48.I80F48
    prices: prices.Prices
    has_open_orders: bool
    has_open_fills: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "PerpInfo":
        return cls(
            perp_market_index=obj.perp_market_index,
            maint_base_asset_weight=i80f48.I80F48.from_decoded(
                obj.maint_base_asset_weight
            ),
            init_base_asset_weight=i80f48.I80F48.from_decoded(
                obj.init_base_asset_weight
            ),
            maint_base_liab_weight=i80f48.I80F48.from_decoded(
                obj.maint_base_liab_weight
            ),
            init_base_liab_weight=i80f48.I80F48.from_decoded(obj.init_base_liab_weight),
            maint_overall_asset_weight=i80f48.I80F48.from_decoded(
                obj.maint_overall_asset_weight
            ),
            init_overall_asset_weight=i80f48.I80F48.from_decoded(
                obj.init_overall_asset_weight
            ),
            base_lot_size=obj.base_lot_size,
            base_lots=obj.base_lots,
            bids_base_lots=obj.bids_base_lots,
            asks_base_lots=obj.asks_base_lots,
            quote=i80f48.I80F48.from_decoded(obj.quote),
            prices=prices.Prices.from_decoded(obj.prices),
            has_open_orders=obj.has_open_orders,
            has_open_fills=obj.has_open_fills,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "perp_market_index": self.perp_market_index,
            "maint_base_asset_weight": self.maint_base_asset_weight.to_encodable(),
            "init_base_asset_weight": self.init_base_asset_weight.to_encodable(),
            "maint_base_liab_weight": self.maint_base_liab_weight.to_encodable(),
            "init_base_liab_weight": self.init_base_liab_weight.to_encodable(),
            "maint_overall_asset_weight": self.maint_overall_asset_weight.to_encodable(),
            "init_overall_asset_weight": self.init_overall_asset_weight.to_encodable(),
            "base_lot_size": self.base_lot_size,
            "base_lots": self.base_lots,
            "bids_base_lots": self.bids_base_lots,
            "asks_base_lots": self.asks_base_lots,
            "quote": self.quote.to_encodable(),
            "prices": self.prices.to_encodable(),
            "has_open_orders": self.has_open_orders,
            "has_open_fills": self.has_open_fills,
        }

    def to_json(self) -> PerpInfoJSON:
        return {
            "perp_market_index": self.perp_market_index,
            "maint_base_asset_weight": self.maint_base_asset_weight.to_json(),
            "init_base_asset_weight": self.init_base_asset_weight.to_json(),
            "maint_base_liab_weight": self.maint_base_liab_weight.to_json(),
            "init_base_liab_weight": self.init_base_liab_weight.to_json(),
            "maint_overall_asset_weight": self.maint_overall_asset_weight.to_json(),
            "init_overall_asset_weight": self.init_overall_asset_weight.to_json(),
            "base_lot_size": self.base_lot_size,
            "base_lots": self.base_lots,
            "bids_base_lots": self.bids_base_lots,
            "asks_base_lots": self.asks_base_lots,
            "quote": self.quote.to_json(),
            "prices": self.prices.to_json(),
            "has_open_orders": self.has_open_orders,
            "has_open_fills": self.has_open_fills,
        }

    @classmethod
    def from_json(cls, obj: PerpInfoJSON) -> "PerpInfo":
        return cls(
            perp_market_index=obj["perp_market_index"],
            maint_base_asset_weight=i80f48.I80F48.from_json(
                obj["maint_base_asset_weight"]
            ),
            init_base_asset_weight=i80f48.I80F48.from_json(
                obj["init_base_asset_weight"]
            ),
            maint_base_liab_weight=i80f48.I80F48.from_json(
                obj["maint_base_liab_weight"]
            ),
            init_base_liab_weight=i80f48.I80F48.from_json(obj["init_base_liab_weight"]),
            maint_overall_asset_weight=i80f48.I80F48.from_json(
                obj["maint_overall_asset_weight"]
            ),
            init_overall_asset_weight=i80f48.I80F48.from_json(
                obj["init_overall_asset_weight"]
            ),
            base_lot_size=obj["base_lot_size"],
            base_lots=obj["base_lots"],
            bids_base_lots=obj["bids_base_lots"],
            asks_base_lots=obj["asks_base_lots"],
            quote=i80f48.I80F48.from_json(obj["quote"]),
            prices=prices.Prices.from_json(obj["prices"]),
            has_open_orders=obj["has_open_orders"],
            has_open_fills=obj["has_open_fills"],
        )



import itertools
import sys
import time
import typing
from dataclasses import dataclass
from decimal import Decimal

from solana.publickey import PublicKey

from mango_explorer_v4.accounts.book_side import BookSide
from mango_explorer_v4.accounts.perp_market import PerpMarket
from mango_explorer_v4.constants import RUST_U64_MAX
from mango_explorer_v4.types.inner_node import InnerNode
from mango_explorer_v4.types.leaf_node import LeafNode
from mango_explorer_v4.types.order_tree_root import OrderTreeRoot
from mango_explorer_v4.helpers.perp_market import PerpMarketHelper


@dataclass
class BookSideItem:
    seq_num: int
    order_id: int
    owner: PublicKey
    open_orders_slot: int
    fee_tier: 0
    ui_price: float
    price: int
    ui_size: float
    size: int
    side: typing.Literal['bids', 'asks']
    timestamp: int
    expiry_timestamp: int
    perp_market_index: int
    is_expired: bool
    is_oracle_pegged: bool
    oracle_pegged_properties: typing.Any


@dataclass
class BookSideItems:
    side: typing.Literal['bids', 'asks']
    book_side: BookSide
    perp_market: PerpMarket
    oracle_price: float

    def __iter__(self):
        def entries(order_tree_root: OrderTreeRoot, is_oracle_pegged: bool):
            if order_tree_root.leaf_count == 0:
                return

            stack = [order_tree_root.maybe_node]

            [left, right] = [1, 0] if self.side == 'bids' else [0, 1]

            now = int(time.time())

            while len(stack) > 0:
                index = stack.pop()

                node = self.book_side.nodes.nodes[index]

                match node.tag:
                    case 1:
                        inner_node = InnerNode.layout.parse(bytes([1] + node.data))

                        stack.extend([inner_node.children[right], inner_node.children[left]])
                    case 2:
                        leaf_node: LeafNode = LeafNode.layout.parse(bytes([2] + node.data))

                        if is_oracle_pegged: # TODO: This won't change - no need to evaluate more than once
                            price_data = leaf_node.key >> 64

                            price_offset = price_data - (1 << 63)

                            price_lots = PerpMarketHelper.ui_price_to_lots(self.perp_market, self.oracle_price) + price_offset

                            is_invalid = {
                                'bids': price_lots > leaf_node.peg_limit,
                                'asks': leaf_node.peg_limit > price_lots
                            }[self.side]

                            oracle_pegged_properties = {
                                'is_invalid': is_invalid,
                                'price_offset': price_offset,
                                'ui_price_offset': PerpMarketHelper.price_lots_to_ui(self.perp_market, price_offset),
                                'peg_limit': leaf_node.peg_limit,
                                'ui_peg_limit': PerpMarketHelper.price_lots_to_ui(self.perp_market, leaf_node.peg_limit)
                            }
                        else:
                            price_lots = leaf_node.key >> 64

                            oracle_pegged_properties = None

                        expiry_timestamp = leaf_node.timestamp + leaf_node.time_in_force if leaf_node.time_in_force else sys.maxsize

                        is_expired = now > expiry_timestamp

                        yield BookSideItem(
                            {
                                'bids': RUST_U64_MAX - (leaf_node.key & ((1 << 64) - 1)),
                                'asks': leaf_node.key & ((1 << 64) - 1)
                            }[self.side],
                            leaf_node.key,
                            leaf_node.owner,
                            leaf_node.owner_slot,
                            0,
                            # float(Decimal(price_lots) * Decimal(self.perp_market.price_lots_to_ui_converter)),
                            float(PerpMarketHelper.price_lots_to_ui(self.perp_market, price_lots)),
                            price_lots,
                            float(PerpMarketHelper.base_lots_to_ui(self.perp_market, leaf_node.quantity)),
                            leaf_node.quantity,
                            self.side,
                            leaf_node.timestamp,
                            expiry_timestamp,
                            self.perp_market.perp_market_index,
                            is_expired,
                            is_oracle_pegged,
                            oracle_pegged_properties
                        )

        fixed_items = entries(self.book_side.roots[0], False)

        oracle_pegged_items = entries(self.book_side.roots[1], True)

        def is_better(side: typing.Literal['bids', 'asks'], a: BookSideItem, b: BookSideItem):
            # TODO: This can be comparison operators instead
            if a.price == b.price:
                return a.seq_num < b.seq_num
                # ^ If prices are equal, prefer the oldest created
            else:
                return {
                    'bids': a.price > b.price,
                    'asks': a.price < b.price
                }[side]

        fixed_item = next(fixed_items, None)

        oracle_pegged_item = next(oracle_pegged_items, None)

        while fixed_item or oracle_pegged_item:
            if fixed_item and oracle_pegged_item:
                if is_better(self.side, fixed_item, oracle_pegged_item):
                    yield fixed_item; fixed_item = next(fixed_items, None)
                else:
                    yield oracle_pegged_item; oracle_pegged_item = next(oracle_pegged_items, None)
            elif fixed_item and not oracle_pegged_item:
                yield fixed_item; fixed_item = next(fixed_items, None)
            elif not fixed_item and oracle_pegged_item:
                yield oracle_pegged_item; oracle_pegged_item = next(oracle_pegged_items, None)

    def l2(self):
        orders = []

        for key, groups in itertools.groupby([[order.ui_price, order.ui_size] for order in self], lambda order: order[0]):
            orders.append([key, float(sum([Decimal(str(size)) for price, size in groups]))])

        return orders

    def impact_price(self, impact_quantity: float):
        accum = 0

        for price, size in self.l2():
            accum += size

            if size >= impact_quantity:
                return price

        return None



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class AccountEditArgs(typing.TypedDict):
    name_opt: typing.Optional[str]
    delegate_opt: typing.Optional[PublicKey]


layout = borsh.CStruct(
    "name_opt" / borsh.Option(borsh.String), "delegate_opt" / borsh.Option(BorshPubkey)
)


class AccountEditAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey


def account_edit(
    args: AccountEditArgs,
    accounts: AccountEditAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xba\xd3\xcd\xb7s]\x18\xa1"
    encoded_args = layout.build(
        {
            "name_opt": args["name_opt"],
            "delegate_opt": args["delegate_opt"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.sysvar import SYSVAR_RENT_PUBKEY
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class TokenRegisterArgs(typing.TypedDict):
    token_index: int
    name: str
    oracle_config: types.oracle_config_params.OracleConfigParams
    interest_rate_params: types.interest_rate_params.InterestRateParams
    loan_fee_rate: float
    loan_origination_fee_rate: float
    maint_asset_weight: float
    init_asset_weight: float
    maint_liab_weight: float
    init_liab_weight: float
    liquidation_fee: float
    min_vault_to_deposits_ratio: float
    net_borrow_limit_window_size_ts: int
    net_borrow_limit_per_window_quote: int


layout = borsh.CStruct(
    "token_index" / borsh.U16,
    "name" / borsh.String,
    "oracle_config" / types.oracle_config_params.OracleConfigParams.layout,
    "interest_rate_params" / types.interest_rate_params.InterestRateParams.layout,
    "loan_fee_rate" / borsh.F32,
    "loan_origination_fee_rate" / borsh.F32,
    "maint_asset_weight" / borsh.F32,
    "init_asset_weight" / borsh.F32,
    "maint_liab_weight" / borsh.F32,
    "init_liab_weight" / borsh.F32,
    "liquidation_fee" / borsh.F32,
    "min_vault_to_deposits_ratio" / borsh.F64,
    "net_borrow_limit_window_size_ts" / borsh.U64,
    "net_borrow_limit_per_window_quote" / borsh.I64,
)


class TokenRegisterAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    mint: PublicKey
    bank: PublicKey
    vault: PublicKey
    mint_info: PublicKey
    oracle: PublicKey
    payer: PublicKey


def token_register(
    args: TokenRegisterArgs,
    accounts: TokenRegisterAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint_info"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYSVAR_RENT_PUBKEY, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"m\x1c\x87:\xa2\xd6q&"
    encoded_args = layout.build(
        {
            "token_index": args["token_index"],
            "name": args["name"],
            "oracle_config": args["oracle_config"].to_encodable(),
            "interest_rate_params": args["interest_rate_params"].to_encodable(),
            "loan_fee_rate": args["loan_fee_rate"],
            "loan_origination_fee_rate": args["loan_origination_fee_rate"],
            "maint_asset_weight": args["maint_asset_weight"],
            "init_asset_weight": args["init_asset_weight"],
            "maint_liab_weight": args["maint_liab_weight"],
            "init_liab_weight": args["init_liab_weight"],
            "liquidation_fee": args["liquidation_fee"],
            "min_vault_to_deposits_ratio": args["min_vault_to_deposits_ratio"],
            "net_borrow_limit_window_size_ts": args["net_borrow_limit_window_size_ts"],
            "net_borrow_limit_per_window_quote": args[
                "net_borrow_limit_per_window_quote"
            ],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID
from .. import types


class PerpMarketJSON(typing.TypedDict):
    group: str
    settle_token_index: int
    perp_market_index: int
    blocked1: int
    group_insurance_fund: int
    bump: int
    base_decimals: int
    name: list[int]
    bids: str
    asks: str
    event_queue: str
    oracle: str
    oracle_config: types.oracle_config.OracleConfigJSON
    stable_price_model: types.stable_price_model.StablePriceModelJSON
    quote_lot_size: int
    base_lot_size: int
    maint_base_asset_weight: types.i80f48.I80F48JSON
    init_base_asset_weight: types.i80f48.I80F48JSON
    maint_base_liab_weight: types.i80f48.I80F48JSON
    init_base_liab_weight: types.i80f48.I80F48JSON
    open_interest: int
    seq_num: int
    registration_time: int
    min_funding: types.i80f48.I80F48JSON
    max_funding: types.i80f48.I80F48JSON
    impact_quantity: int
    long_funding: types.i80f48.I80F48JSON
    short_funding: types.i80f48.I80F48JSON
    funding_last_updated: int
    base_liquidation_fee: types.i80f48.I80F48JSON
    maker_fee: types.i80f48.I80F48JSON
    taker_fee: types.i80f48.I80F48JSON
    fees_accrued: types.i80f48.I80F48JSON
    fees_settled: types.i80f48.I80F48JSON
    fee_penalty: float
    settle_fee_flat: float
    settle_fee_amount_threshold: float
    settle_fee_fraction_low_health: float
    settle_pnl_limit_factor: float
    padding3: list[int]
    settle_pnl_limit_window_size_ts: int
    reduce_only: int
    padding4: list[int]
    maint_overall_asset_weight: types.i80f48.I80F48JSON
    init_overall_asset_weight: types.i80f48.I80F48JSON
    positive_pnl_liquidation_fee: types.i80f48.I80F48JSON
    reserved: list[int]


@dataclass
class PerpMarket:
    discriminator: typing.ClassVar = b"\n\xdf\x0c,k\xf57\xf7"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "settle_token_index" / borsh.U16,
        "perp_market_index" / borsh.U16,
        "blocked1" / borsh.U8,
        "group_insurance_fund" / borsh.U8,
        "bump" / borsh.U8,
        "base_decimals" / borsh.U8,
        "name" / borsh.U8[16],
        "bids" / BorshPubkey,
        "asks" / BorshPubkey,
        "event_queue" / BorshPubkey,
        "oracle" / BorshPubkey,
        "oracle_config" / types.oracle_config.OracleConfig.layout,
        "stable_price_model" / types.stable_price_model.StablePriceModel.layout,
        "quote_lot_size" / borsh.I64,
        "base_lot_size" / borsh.I64,
        "maint_base_asset_weight" / types.i80f48.I80F48.layout,
        "init_base_asset_weight" / types.i80f48.I80F48.layout,
        "maint_base_liab_weight" / types.i80f48.I80F48.layout,
        "init_base_liab_weight" / types.i80f48.I80F48.layout,
        "open_interest" / borsh.I64,
        "seq_num" / borsh.U64,
        "registration_time" / borsh.U64,
        "min_funding" / types.i80f48.I80F48.layout,
        "max_funding" / types.i80f48.I80F48.layout,
        "impact_quantity" / borsh.I64,
        "long_funding" / types.i80f48.I80F48.layout,
        "short_funding" / types.i80f48.I80F48.layout,
        "funding_last_updated" / borsh.U64,
        "base_liquidation_fee" / types.i80f48.I80F48.layout,
        "maker_fee" / types.i80f48.I80F48.layout,
        "taker_fee" / types.i80f48.I80F48.layout,
        "fees_accrued" / types.i80f48.I80F48.layout,
        "fees_settled" / types.i80f48.I80F48.layout,
        "fee_penalty" / borsh.F32,
        "settle_fee_flat" / borsh.F32,
        "settle_fee_amount_threshold" / borsh.F32,
        "settle_fee_fraction_low_health" / borsh.F32,
        "settle_pnl_limit_factor" / borsh.F32,
        "padding3" / borsh.U8[4],
        "settle_pnl_limit_window_size_ts" / borsh.U64,
        "reduce_only" / borsh.U8,
        "padding4" / borsh.U8[7],
        "maint_overall_asset_weight" / types.i80f48.I80F48.layout,
        "init_overall_asset_weight" / types.i80f48.I80F48.layout,
        "positive_pnl_liquidation_fee" / types.i80f48.I80F48.layout,
        "reserved" / borsh.U8[1888],
    )
    group: PublicKey
    settle_token_index: int
    perp_market_index: int
    blocked1: int
    group_insurance_fund: int
    bump: int
    base_decimals: int
    name: list[int]
    bids: PublicKey
    asks: PublicKey
    event_queue: PublicKey
    oracle: PublicKey
    oracle_config: types.oracle_config.OracleConfig
    stable_price_model: types.stable_price_model.StablePriceModel
    quote_lot_size: int
    base_lot_size: int
    maint_base_asset_weight: types.i80f48.I80F48
    init_base_asset_weight: types.i80f48.I80F48
    maint_base_liab_weight: types.i80f48.I80F48
    init_base_liab_weight: types.i80f48.I80F48
    open_interest: int
    seq_num: int
    registration_time: int
    min_funding: types.i80f48.I80F48
    max_funding: types.i80f48.I80F48
    impact_quantity: int
    long_funding: types.i80f48.I80F48
    short_funding: types.i80f48.I80F48
    funding_last_updated: int
    base_liquidation_fee: types.i80f48.I80F48
    maker_fee: types.i80f48.I80F48
    taker_fee: types.i80f48.I80F48
    fees_accrued: types.i80f48.I80F48
    fees_settled: types.i80f48.I80F48
    fee_penalty: float
    settle_fee_flat: float
    settle_fee_amount_threshold: float
    settle_fee_fraction_low_health: float
    settle_pnl_limit_factor: float
    padding3: list[int]
    settle_pnl_limit_window_size_ts: int
    reduce_only: int
    padding4: list[int]
    maint_overall_asset_weight: types.i80f48.I80F48
    init_overall_asset_weight: types.i80f48.I80F48
    positive_pnl_liquidation_fee: types.i80f48.I80F48
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["PerpMarket"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["PerpMarket"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["PerpMarket"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "PerpMarket":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = PerpMarket.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            group=dec.group,
            settle_token_index=dec.settle_token_index,
            perp_market_index=dec.perp_market_index,
            blocked1=dec.blocked1,
            group_insurance_fund=dec.group_insurance_fund,
            bump=dec.bump,
            base_decimals=dec.base_decimals,
            name=dec.name,
            bids=dec.bids,
            asks=dec.asks,
            event_queue=dec.event_queue,
            oracle=dec.oracle,
            oracle_config=types.oracle_config.OracleConfig.from_decoded(
                dec.oracle_config
            ),
            stable_price_model=types.stable_price_model.StablePriceModel.from_decoded(
                dec.stable_price_model
            ),
            quote_lot_size=dec.quote_lot_size,
            base_lot_size=dec.base_lot_size,
            maint_base_asset_weight=types.i80f48.I80F48.from_decoded(
                dec.maint_base_asset_weight
            ),
            init_base_asset_weight=types.i80f48.I80F48.from_decoded(
                dec.init_base_asset_weight
            ),
            maint_base_liab_weight=types.i80f48.I80F48.from_decoded(
                dec.maint_base_liab_weight
            ),
            init_base_liab_weight=types.i80f48.I80F48.from_decoded(
                dec.init_base_liab_weight
            ),
            open_interest=dec.open_interest,
            seq_num=dec.seq_num,
            registration_time=dec.registration_time,
            min_funding=types.i80f48.I80F48.from_decoded(dec.min_funding),
            max_funding=types.i80f48.I80F48.from_decoded(dec.max_funding),
            impact_quantity=dec.impact_quantity,
            long_funding=types.i80f48.I80F48.from_decoded(dec.long_funding),
            short_funding=types.i80f48.I80F48.from_decoded(dec.short_funding),
            funding_last_updated=dec.funding_last_updated,
            base_liquidation_fee=types.i80f48.I80F48.from_decoded(
                dec.base_liquidation_fee
            ),
            maker_fee=types.i80f48.I80F48.from_decoded(dec.maker_fee),
            taker_fee=types.i80f48.I80F48.from_decoded(dec.taker_fee),
            fees_accrued=types.i80f48.I80F48.from_decoded(dec.fees_accrued),
            fees_settled=types.i80f48.I80F48.from_decoded(dec.fees_settled),
            fee_penalty=dec.fee_penalty,
            settle_fee_flat=dec.settle_fee_flat,
            settle_fee_amount_threshold=dec.settle_fee_amount_threshold,
            settle_fee_fraction_low_health=dec.settle_fee_fraction_low_health,
            settle_pnl_limit_factor=dec.settle_pnl_limit_factor,
            padding3=dec.padding3,
            settle_pnl_limit_window_size_ts=dec.settle_pnl_limit_window_size_ts,
            reduce_only=dec.reduce_only,
            padding4=dec.padding4,
            maint_overall_asset_weight=types.i80f48.I80F48.from_decoded(
                dec.maint_overall_asset_weight
            ),
            init_overall_asset_weight=types.i80f48.I80F48.from_decoded(
                dec.init_overall_asset_weight
            ),
            positive_pnl_liquidation_fee=types.i80f48.I80F48.from_decoded(
                dec.positive_pnl_liquidation_fee
            ),
            reserved=dec.reserved,
        )

    def to_json(self) -> PerpMarketJSON:
        return {
            "group": str(self.group),
            "settle_token_index": self.settle_token_index,
            "perp_market_index": self.perp_market_index,
            "blocked1": self.blocked1,
            "group_insurance_fund": self.group_insurance_fund,
            "bump": self.bump,
            "base_decimals": self.base_decimals,
            "name": self.name,
            "bids": str(self.bids),
            "asks": str(self.asks),
            "event_queue": str(self.event_queue),
            "oracle": str(self.oracle),
            "oracle_config": self.oracle_config.to_json(),
            "stable_price_model": self.stable_price_model.to_json(),
            "quote_lot_size": self.quote_lot_size,
            "base_lot_size": self.base_lot_size,
            "maint_base_asset_weight": self.maint_base_asset_weight.to_json(),
            "init_base_asset_weight": self.init_base_asset_weight.to_json(),
            "maint_base_liab_weight": self.maint_base_liab_weight.to_json(),
            "init_base_liab_weight": self.init_base_liab_weight.to_json(),
            "open_interest": self.open_interest,
            "seq_num": self.seq_num,
            "registration_time": self.registration_time,
            "min_funding": self.min_funding.to_json(),
            "max_funding": self.max_funding.to_json(),
            "impact_quantity": self.impact_quantity,
            "long_funding": self.long_funding.to_json(),
            "short_funding": self.short_funding.to_json(),
            "funding_last_updated": self.funding_last_updated,
            "base_liquidation_fee": self.base_liquidation_fee.to_json(),
            "maker_fee": self.maker_fee.to_json(),
            "taker_fee": self.taker_fee.to_json(),
            "fees_accrued": self.fees_accrued.to_json(),
            "fees_settled": self.fees_settled.to_json(),
            "fee_penalty": self.fee_penalty,
            "settle_fee_flat": self.settle_fee_flat,
            "settle_fee_amount_threshold": self.settle_fee_amount_threshold,
            "settle_fee_fraction_low_health": self.settle_fee_fraction_low_health,
            "settle_pnl_limit_factor": self.settle_pnl_limit_factor,
            "padding3": self.padding3,
            "settle_pnl_limit_window_size_ts": self.settle_pnl_limit_window_size_ts,
            "reduce_only": self.reduce_only,
            "padding4": self.padding4,
            "maint_overall_asset_weight": self.maint_overall_asset_weight.to_json(),
            "init_overall_asset_weight": self.init_overall_asset_weight.to_json(),
            "positive_pnl_liquidation_fee": self.positive_pnl_liquidation_fee.to_json(),
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: PerpMarketJSON) -> "PerpMarket":
        return cls(
            group=PublicKey(obj["group"]),
            settle_token_index=obj["settle_token_index"],
            perp_market_index=obj["perp_market_index"],
            blocked1=obj["blocked1"],
            group_insurance_fund=obj["group_insurance_fund"],
            bump=obj["bump"],
            base_decimals=obj["base_decimals"],
            name=obj["name"],
            bids=PublicKey(obj["bids"]),
            asks=PublicKey(obj["asks"]),
            event_queue=PublicKey(obj["event_queue"]),
            oracle=PublicKey(obj["oracle"]),
            oracle_config=types.oracle_config.OracleConfig.from_json(
                obj["oracle_config"]
            ),
            stable_price_model=types.stable_price_model.StablePriceModel.from_json(
                obj["stable_price_model"]
            ),
            quote_lot_size=obj["quote_lot_size"],
            base_lot_size=obj["base_lot_size"],
            maint_base_asset_weight=types.i80f48.I80F48.from_json(
                obj["maint_base_asset_weight"]
            ),
            init_base_asset_weight=types.i80f48.I80F48.from_json(
                obj["init_base_asset_weight"]
            ),
            maint_base_liab_weight=types.i80f48.I80F48.from_json(
                obj["maint_base_liab_weight"]
            ),
            init_base_liab_weight=types.i80f48.I80F48.from_json(
                obj["init_base_liab_weight"]
            ),
            open_interest=obj["open_interest"],
            seq_num=obj["seq_num"],
            registration_time=obj["registration_time"],
            min_funding=types.i80f48.I80F48.from_json(obj["min_funding"]),
            max_funding=types.i80f48.I80F48.from_json(obj["max_funding"]),
            impact_quantity=obj["impact_quantity"],
            long_funding=types.i80f48.I80F48.from_json(obj["long_funding"]),
            short_funding=types.i80f48.I80F48.from_json(obj["short_funding"]),
            funding_last_updated=obj["funding_last_updated"],
            base_liquidation_fee=types.i80f48.I80F48.from_json(
                obj["base_liquidation_fee"]
            ),
            maker_fee=types.i80f48.I80F48.from_json(obj["maker_fee"]),
            taker_fee=types.i80f48.I80F48.from_json(obj["taker_fee"]),
            fees_accrued=types.i80f48.I80F48.from_json(obj["fees_accrued"]),
            fees_settled=types.i80f48.I80F48.from_json(obj["fees_settled"]),
            fee_penalty=obj["fee_penalty"],
            settle_fee_flat=obj["settle_fee_flat"],
            settle_fee_amount_threshold=obj["settle_fee_amount_threshold"],
            settle_fee_fraction_low_health=obj["settle_fee_fraction_low_health"],
            settle_pnl_limit_factor=obj["settle_pnl_limit_factor"],
            padding3=obj["padding3"],
            settle_pnl_limit_window_size_ts=obj["settle_pnl_limit_window_size_ts"],
            reduce_only=obj["reduce_only"],
            padding4=obj["padding4"],
            maint_overall_asset_weight=types.i80f48.I80F48.from_json(
                obj["maint_overall_asset_weight"]
            ),
            init_overall_asset_weight=types.i80f48.I80F48.from_json(
                obj["init_overall_asset_weight"]
            ),
            positive_pnl_liquidation_fee=types.i80f48.I80F48.from_json(
                obj["positive_pnl_liquidation_fee"]
            ),
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class EventQueueHeaderJSON(typing.TypedDict):
    head: int
    count: int
    seq_num: int


@dataclass
class EventQueueHeader:
    layout: typing.ClassVar = borsh.CStruct(
        "head" / borsh.U32, "count" / borsh.U32, "seq_num" / borsh.U64
    )
    head: int
    count: int
    seq_num: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "EventQueueHeader":
        return cls(head=obj.head, count=obj.count, seq_num=obj.seq_num)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"head": self.head, "count": self.count, "seq_num": self.seq_num}

    def to_json(self) -> EventQueueHeaderJSON:
        return {"head": self.head, "count": self.count, "seq_num": self.seq_num}

    @classmethod
    def from_json(cls, obj: EventQueueHeaderJSON) -> "EventQueueHeader":
        return cls(head=obj["head"], count=obj["count"], seq_num=obj["seq_num"])



from .group_create import group_create, GroupCreateArgs, GroupCreateAccounts
from .group_edit import group_edit, GroupEditArgs, GroupEditAccounts
from .ix_gate_set import ix_gate_set, IxGateSetArgs, IxGateSetAccounts
from .group_close import group_close, GroupCloseAccounts
from .token_register import token_register, TokenRegisterArgs, TokenRegisterAccounts
from .token_register_trustless import (
    token_register_trustless,
    TokenRegisterTrustlessArgs,
    TokenRegisterTrustlessAccounts,
)
from .token_edit import token_edit, TokenEditArgs, TokenEditAccounts
from .token_add_bank import token_add_bank, TokenAddBankArgs, TokenAddBankAccounts
from .token_deregister import token_deregister, TokenDeregisterAccounts
from .token_update_index_and_rate import (
    token_update_index_and_rate,
    TokenUpdateIndexAndRateAccounts,
)
from .account_create import account_create, AccountCreateArgs, AccountCreateAccounts
from .account_expand import account_expand, AccountExpandArgs, AccountExpandAccounts
from .account_edit import account_edit, AccountEditArgs, AccountEditAccounts
from .account_toggle_freeze import (
    account_toggle_freeze,
    AccountToggleFreezeArgs,
    AccountToggleFreezeAccounts,
)
from .account_close import account_close, AccountCloseArgs, AccountCloseAccounts
from .account_buyback_fees_with_mngo import (
    account_buyback_fees_with_mngo,
    AccountBuybackFeesWithMngoArgs,
    AccountBuybackFeesWithMngoAccounts,
)
from .stub_oracle_create import (
    stub_oracle_create,
    StubOracleCreateArgs,
    StubOracleCreateAccounts,
)
from .stub_oracle_close import stub_oracle_close, StubOracleCloseAccounts
from .stub_oracle_set import stub_oracle_set, StubOracleSetArgs, StubOracleSetAccounts
from .token_deposit import token_deposit, TokenDepositArgs, TokenDepositAccounts
from .token_deposit_into_existing import (
    token_deposit_into_existing,
    TokenDepositIntoExistingArgs,
    TokenDepositIntoExistingAccounts,
)
from .token_withdraw import token_withdraw, TokenWithdrawArgs, TokenWithdrawAccounts
from .flash_loan_begin import (
    flash_loan_begin,
    FlashLoanBeginArgs,
    FlashLoanBeginAccounts,
)
from .flash_loan_end import flash_loan_end, FlashLoanEndArgs, FlashLoanEndAccounts
from .health_region_begin import health_region_begin, HealthRegionBeginAccounts
from .health_region_end import health_region_end, HealthRegionEndAccounts
from .serum3_register_market import (
    serum3_register_market,
    Serum3RegisterMarketArgs,
    Serum3RegisterMarketAccounts,
)
from .serum3_edit_market import (
    serum3_edit_market,
    Serum3EditMarketArgs,
    Serum3EditMarketAccounts,
)
from .serum3_deregister_market import (
    serum3_deregister_market,
    Serum3DeregisterMarketAccounts,
)
from .serum3_create_open_orders import (
    serum3_create_open_orders,
    Serum3CreateOpenOrdersAccounts,
)
from .serum3_close_open_orders import (
    serum3_close_open_orders,
    Serum3CloseOpenOrdersAccounts,
)
from .serum3_place_order import (
    serum3_place_order,
    Serum3PlaceOrderArgs,
    Serum3PlaceOrderAccounts,
)
from .serum3_cancel_order import (
    serum3_cancel_order,
    Serum3CancelOrderArgs,
    Serum3CancelOrderAccounts,
)
from .serum3_cancel_all_orders import (
    serum3_cancel_all_orders,
    Serum3CancelAllOrdersArgs,
    Serum3CancelAllOrdersAccounts,
)
from .serum3_settle_funds import serum3_settle_funds, Serum3SettleFundsAccounts
from .serum3_settle_funds_v2 import (
    serum3_settle_funds_v2,
    Serum3SettleFundsV2Args,
    Serum3SettleFundsV2Accounts,
)
from .serum3_liq_force_cancel_orders import (
    serum3_liq_force_cancel_orders,
    Serum3LiqForceCancelOrdersArgs,
    Serum3LiqForceCancelOrdersAccounts,
)
from .liq_token_with_token import (
    liq_token_with_token,
    LiqTokenWithTokenArgs,
    LiqTokenWithTokenAccounts,
)
from .liq_token_bankruptcy import (
    liq_token_bankruptcy,
    LiqTokenBankruptcyArgs,
    LiqTokenBankruptcyAccounts,
)
from .token_liq_with_token import (
    token_liq_with_token,
    TokenLiqWithTokenArgs,
    TokenLiqWithTokenAccounts,
)
from .token_liq_bankruptcy import (
    token_liq_bankruptcy,
    TokenLiqBankruptcyArgs,
    TokenLiqBankruptcyAccounts,
)
from .perp_create_market import (
    perp_create_market,
    PerpCreateMarketArgs,
    PerpCreateMarketAccounts,
)
from .perp_edit_market import (
    perp_edit_market,
    PerpEditMarketArgs,
    PerpEditMarketAccounts,
)
from .perp_close_market import perp_close_market, PerpCloseMarketAccounts
from .perp_deactivate_position import (
    perp_deactivate_position,
    PerpDeactivatePositionAccounts,
)
from .perp_place_order import (
    perp_place_order,
    PerpPlaceOrderArgs,
    PerpPlaceOrderAccounts,
)
from .perp_place_order_pegged import (
    perp_place_order_pegged,
    PerpPlaceOrderPeggedArgs,
    PerpPlaceOrderPeggedAccounts,
)
from .perp_cancel_order import (
    perp_cancel_order,
    PerpCancelOrderArgs,
    PerpCancelOrderAccounts,
)
from .perp_cancel_order_by_client_order_id import (
    perp_cancel_order_by_client_order_id,
    PerpCancelOrderByClientOrderIdArgs,
    PerpCancelOrderByClientOrderIdAccounts,
)
from .perp_cancel_all_orders import (
    perp_cancel_all_orders,
    PerpCancelAllOrdersArgs,
    PerpCancelAllOrdersAccounts,
)
from .perp_cancel_all_orders_by_side import (
    perp_cancel_all_orders_by_side,
    PerpCancelAllOrdersBySideArgs,
    PerpCancelAllOrdersBySideAccounts,
)
from .perp_consume_events import (
    perp_consume_events,
    PerpConsumeEventsArgs,
    PerpConsumeEventsAccounts,
)
from .perp_update_funding import perp_update_funding, PerpUpdateFundingAccounts
from .perp_settle_pnl import perp_settle_pnl, PerpSettlePnlAccounts
from .perp_settle_fees import (
    perp_settle_fees,
    PerpSettleFeesArgs,
    PerpSettleFeesAccounts,
)
from .perp_liq_base_or_positive_pnl import (
    perp_liq_base_or_positive_pnl,
    PerpLiqBaseOrPositivePnlArgs,
    PerpLiqBaseOrPositivePnlAccounts,
)
from .perp_liq_force_cancel_orders import (
    perp_liq_force_cancel_orders,
    PerpLiqForceCancelOrdersArgs,
    PerpLiqForceCancelOrdersAccounts,
)
from .perp_liq_negative_pnl_or_bankruptcy import (
    perp_liq_negative_pnl_or_bankruptcy,
    PerpLiqNegativePnlOrBankruptcyArgs,
    PerpLiqNegativePnlOrBankruptcyAccounts,
)
from .alt_set import alt_set, AltSetArgs, AltSetAccounts
from .alt_extend import alt_extend, AltExtendArgs, AltExtendAccounts
from .compute_account_data import compute_account_data, ComputeAccountDataAccounts
from .benchmark import benchmark



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class Serum3PlaceOrderArgs(typing.TypedDict):
    side: types.serum3_side.Serum3SideKind
    limit_price: int
    max_base_qty: int
    max_native_quote_qty_including_fees: int
    self_trade_behavior: types.serum3_self_trade_behavior.Serum3SelfTradeBehaviorKind
    order_type: types.serum3_order_type.Serum3OrderTypeKind
    client_order_id: int
    limit: int


layout = borsh.CStruct(
    "side" / types.serum3_side.layout,
    "limit_price" / borsh.U64,
    "max_base_qty" / borsh.U64,
    "max_native_quote_qty_including_fees" / borsh.U64,
    "self_trade_behavior" / types.serum3_self_trade_behavior.layout,
    "order_type" / types.serum3_order_type.layout,
    "client_order_id" / borsh.U64,
    "limit" / borsh.U16,
)


class Serum3PlaceOrderAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    open_orders: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_bids: PublicKey
    market_asks: PublicKey
    market_event_queue: PublicKey
    market_request_queue: PublicKey
    market_base_vault: PublicKey
    market_quote_vault: PublicKey
    market_vault_signer: PublicKey
    payer_bank: PublicKey
    payer_vault: PublicKey
    payer_oracle: PublicKey


def serum3_place_order(
    args: Serum3PlaceOrderArgs,
    accounts: Serum3PlaceOrderAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["serum_market"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_market_external"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["market_bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_asks"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["market_event_queue"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_request_queue"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_base_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_quote_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_vault_signer"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["payer_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["payer_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["payer_oracle"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"a\x1d{\xc7\xe4\x14\xb8\xfc"
    encoded_args = layout.build(
        {
            "side": args["side"].to_encodable(),
            "limit_price": args["limit_price"],
            "max_base_qty": args["max_base_qty"],
            "max_native_quote_qty_including_fees": args[
                "max_native_quote_qty_including_fees"
            ],
            "self_trade_behavior": args["self_trade_behavior"].to_encodable(),
            "order_type": args["order_type"].to_encodable(),
            "client_order_id": args["client_order_id"],
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)
